#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import time

import numpy as np
import pytest

import synnax as sy
from synnax.util.random import random_name


@pytest.mark.framer
@pytest.mark.calculations
class TestCalculatedChannelStreaming:
    def test_basic_calculated_channel_stream(self, client: sy.Synnax):
        """Should correctly create and read from a basic calculated channel using streaming"""
        timestamp_channel = client.channels.create(
            name=random_name(),
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
        )
        src_channels = client.channels.create(
            [
                sy.Channel(
                    name=random_name(),
                    index=timestamp_channel.key,
                    data_type=sy.DataType.FLOAT32,
                ),
                sy.Channel(
                    name=random_name(),
                    index=timestamp_channel.key,
                    data_type=sy.DataType.FLOAT32,
                ),
            ]
        )
        calc_channel = client.channels.create(
            name=random_name(),
            expression=f"return {src_channels[0].name} + {src_channels[1].name}",
        )
        start = sy.TimeStamp.now()
        value = np.array(
            [2.0],
            dtype=np.float32,
        )
        with client.open_streamer(calc_channel.key) as streamer:
            with client.open_writer(
                start,
                [timestamp_channel.key, src_channels[0].key, src_channels[1].key],
            ) as writer:
                alignment_hwm = 0
                for i in range(5):
                    time.sleep(0.01)
                    writer.write(
                        {
                            timestamp_channel.key: sy.TimeStamp.now(),
                            src_channels[0].key: value / 2,
                            src_channels[1].key: value / 2,
                        }
                    )
                    frame = streamer.read(timeout=100)
                    assert len(frame.channels) == 1
                    ser = frame[calc_channel.key]
                    assert ser.alignment != 0
                    if i != 0:
                        alignment_hwm += 1
                        assert ser.alignment == alignment_hwm
                    else:
                        alignment_hwm = ser.alignment
                    assert frame is not None
                    assert np.array_equal(frame[calc_channel.key], value)

    def test_stream_passthrough_virtual_channel(self, client: sy.Synnax):
        virt = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            virtual=True,
        )
        calc = client.channels.create(
            name=random_name(),
            expression=f"return {virt.name} * 2",
        )
        start = sy.TimeStamp.now()
        value = np.array([3], dtype=np.uint8)
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(
                start,
                [virt.key],
            ) as writer:
                writer.write(
                    {
                        virt.key: value,
                    }
                )
                writer.commit()
                frame = streamer.read(timeout=2)
                assert frame is not None
                assert np.array_equal(frame[calc.key], value * 2)

    def test_stream_passthrough_virtual_u8_channel(self, client: sy.Synnax):
        virt = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.UINT8,
            virtual=True,
        )
        calc = client.channels.create(
            name=random_name(),
            expression=f"return {virt.name} * 2",
        )
        start = sy.TimeStamp.now()
        value = np.array([3], dtype=np.uint8)
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(
                start,
                [virt.key],
            ) as writer:
                writer.write(
                    {
                        virt.key: value,
                    }
                )
                writer.commit()
                frame = streamer.read(timeout=2)
                assert frame is not None
                assert np.array_equal(frame[calc.key], value * 2)


@pytest.mark.framer
@pytest.mark.calculations
class TestCalculatedChannelIteration:
    def test_basic_calculated_channel_iterate(self, client: sy.Synnax):
        """Should correctly create and read from a basic calculated channel using iteration"""
        timestamp_channel = client.channels.create(
            name=random_name(),
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
        )
        src_channel = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=timestamp_channel.key,
                data_type=sy.DataType.FLOAT32,
            ),
        )
        calc_channel = client.channels.create(
            name=random_name(),
            expression=f"return {src_channel.name}",
        )
        idx_data_1 = [
            5 * sy.TimeSpan.SECOND,
            6 * sy.TimeSpan.SECOND,
            7 * sy.TimeSpan.SECOND,
        ]
        src_data_1 = [10.0, 20.0, 30.0]
        client.write(
            5 * sy.TimeSpan.SECOND,
            {
                timestamp_channel.key: idx_data_1,
                src_channel.key: src_data_1,
            },
        )
        res = client.read(
            tr=sy.TimeRange.MAX,
            channels=[
                calc_channel.key,
                calc_channel.index,
            ],
        )
        assert len(res.channels) == 2
        ts_ser = res[calc_channel.index]
        data_ser = res[calc_channel.key]
        assert len(ts_ser) == 3
        assert ts_ser.alignment == data_ser.alignment
        assert np.array_equal(ts_ser, np.array(idx_data_1, dtype=ts_ser.data_type.np))
        assert np.array_equal(
            data_ser, np.array(src_data_1, dtype=data_ser.data_type.np)
        )

        idx_data_2 = [
            9 * sy.TimeSpan.SECOND,
            10 * sy.TimeSpan.SECOND,
            11 * sy.TimeSpan.SECOND,
        ]
        src_data_2 = [50.0, 60.0, 70.0]
        client.write(
            9 * sy.TimeSpan.SECOND,
            {
                timestamp_channel.key: idx_data_2,
                src_channel.key: src_data_2,
            },
        )
        res = client.read(
            tr=sy.TimeRange.MAX,
            channels=[calc_channel.key, calc_channel.index],
        )
        assert len(res.channels) == 4
        ts_ser = res[calc_channel.index]
        data_ser = res[calc_channel.key]
        assert len(ts_ser) == 6
        assert ts_ser.alignment == data_ser.alignment
        assert np.array_equal(
            ts_ser,
            np.array(idx_data_1 + idx_data_2, dtype=ts_ser.data_type.np),
        )

        idx_data_0 = [1 * sy.TimeSpan.SECOND, 2 * sy.TimeSpan.SECOND]
        src_data_0 = [0.0, 5.0]
        client.write(
            1 * sy.TimeSpan.SECOND,
            {
                timestamp_channel.key: idx_data_0,
                src_channel.key: src_data_0,
            },
        )
        res = client.read(
            tr=sy.TimeRange.MAX,
            channels=[calc_channel.key, calc_channel.index],
        )
        assert len(res.channels) == 6
        data_ser = res[calc_channel.key]
        ts_ser = res[calc_channel.index]
        assert len(ts_ser) == 8
        assert ts_ser.alignment == data_ser.alignment
        assert np.array_equal(
            ts_ser,
            np.array(idx_data_0 + idx_data_1 + idx_data_2, dtype=ts_ser.data_type.np),
        )
        assert np.array_equal(
            data_ser,
            np.array(src_data_0 + src_data_1 + src_data_2, dtype=data_ser.data_type.np),
        )

    def test_conditional_calculated_channel(self, client: sy.Synnax):
        """Should correctly create and read from a basic calculated channel using iteration"""
        timestamp_channel = client.channels.create(
            name=random_name(),
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
        )
        src_channel = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=timestamp_channel.key,
                data_type=sy.DataType.FLOAT32,
            ),
        )
        calc_channel = client.channels.create(
            name=random_name(),
            expression=f"""
            if ({src_channel.name} > 15) {{
                return 4
            }} else {{
                return 5
            }}
            """,
        )
        idx_data = [
            5 * sy.TimeSpan.SECOND,
            6 * sy.TimeSpan.SECOND,
            7 * sy.TimeSpan.SECOND,
        ]
        src_data = [10.0, 20.0, 30.0]
        client.write(
            5 * sy.TimeSpan.SECOND,
            {
                timestamp_channel.key: idx_data,
                src_channel.key: src_data,
            },
        )
        res = client.read(
            tr=sy.TimeRange.MAX,
            channels=[
                calc_channel.key,
                calc_channel.index,
            ],
        )
        assert len(res.channels) == 2
        ts_ser = res[calc_channel.index]
        data_ser = res[calc_channel.key]
        assert len(ts_ser) == 3
        assert ts_ser.alignment == data_ser.alignment
        assert np.array_equal(ts_ser, np.array(idx_data, dtype=ts_ser.data_type.np))
        assert np.array_equal(
            data_ser, np.array([5, 4, 4], dtype=data_ser.data_type.np)
        )

    def test_calculation_deleted_channel(self, client: sy.Synnax):
        """Should correctly create and read from a basic calculated channel using iteration"""
        timestamp_channel = client.channels.create(
            name=random_name(),
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
        )
        src_channel = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=timestamp_channel.key,
                data_type=sy.DataType.FLOAT32,
            ),
        )
        calc_channel = client.channels.create(
            name=random_name(),
            expression=f"""
              if ({src_channel.name} > 15) {{
                  return 4
              }} else {{
                  return 5
              }}""",
        )
        idx_data = [
            5 * sy.TimeSpan.SECOND,
            6 * sy.TimeSpan.SECOND,
            7 * sy.TimeSpan.SECOND,
        ]
        src_data = [10.0, 20.0, 30.0]
        client.write(
            5 * sy.TimeSpan.SECOND,
            {
                timestamp_channel.key: idx_data,
                src_channel.key: src_data,
            },
        )
        res = client.read(
            tr=sy.TimeRange.MAX,
            channels=[
                calc_channel.key,
                calc_channel.index,
            ],
        )
        assert len(res.channels) == 2
        ts_ser = res[calc_channel.index]
        data_ser = res[calc_channel.key]
        assert len(ts_ser) == 3
        assert ts_ser.alignment == data_ser.alignment
        assert np.array_equal(ts_ser, np.array(idx_data, dtype=ts_ser.data_type.np))
        assert np.array_equal(
            data_ser, np.array([5, 4, 4], dtype=data_ser.data_type.np)
        )
        client.channels.delete(src_channel.key)
        with pytest.raises(Exception, match="undefined symbol"):
            client.read(
                tr=sy.TimeRange.MAX,
                channels=[
                    calc_channel.key,
                    calc_channel.index,
                ],
            )

    def test_nested_calculated_channels_2_level(self, client: sy.Synnax):
        """Should correctly handle 2-level nested calculated channels (C → B → A)"""
        # Create index and base concrete channel
        timestamp_channel = client.channels.create(
            name=random_name(),
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
        )
        sensor_1 = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=timestamp_channel.key,
                data_type=sy.DataType.FLOAT32,
            ),
        )

        # Create B: calculated channel that depends on concrete channel A (sensor_1)
        calc_b = client.channels.create(
            name=random_name(),
            expression=f"return {sensor_1.name} * 2",
        )

        # Create C: calculated channel that depends on calculated channel B
        calc_c = client.channels.create(
            name=random_name(),
            expression=f"return {calc_b.name} + 10",
        )

        # Write data
        idx_data = [
            1 * sy.TimeSpan.SECOND,
            2 * sy.TimeSpan.SECOND,
            3 * sy.TimeSpan.SECOND,
            4 * sy.TimeSpan.SECOND,
            5 * sy.TimeSpan.SECOND,
        ]
        src_data = [1.0, 2.0, 3.0, 4.0, 5.0]
        client.write(
            1 * sy.TimeSpan.SECOND,
            {
                timestamp_channel.key: idx_data,
                sensor_1.key: src_data,
            },
        )

        # Read the top-level calculated channel C
        res = client.read(
            tr=sy.TimeRange.MAX,
            channels=[calc_c.key, calc_c.index],
        )

        # Verify results
        # sensor_1 = [1, 2, 3, 4, 5]
        # calc_b = sensor_1 * 2 = [2, 4, 6, 8, 10]
        # calc_c = calc_b + 10 = [12, 14, 16, 18, 20]

        # Verify requested channels are present and have correct data
        assert calc_c.key in res.channels
        assert calc_c.index in res.channels

        data_ser = res[calc_c.key]
        ts_ser = res[calc_c.index]
        assert len(ts_ser) == 5
        assert np.array_equal(
            data_ser, np.array([12.0, 14.0, 16.0, 18.0, 20.0], dtype=np.float32)
        )
        assert np.array_equal(ts_ser, np.array(idx_data, dtype=ts_ser.data_type.np))

    def test_nested_calculated_channels_3_level(self, client: sy.Synnax):
        """Should correctly handle 3-level nested calculated channels (D → C → B → A)"""
        # Create index and base concrete channel
        timestamp_channel = client.channels.create(
            name=random_name(),
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
        )
        sensor_1 = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=timestamp_channel.key,
                data_type=sy.DataType.FLOAT32,
            ),
        )

        # Create B: depends on sensor_1 (concrete)
        calc_b = client.channels.create(
            name=random_name(),
            expression=f"return {sensor_1.name} * 2",
        )

        # Create C: depends on B (calculated)
        calc_c = client.channels.create(
            name=random_name(),
            expression=f"return {calc_b.name} + 5",
        )

        # Create D: depends on C (calculated)
        calc_d = client.channels.create(
            name=random_name(),
            expression=f"return {calc_c.name} * 3",
        )

        # Write data
        idx_data = [
            1 * sy.TimeSpan.SECOND,
            2 * sy.TimeSpan.SECOND,
            3 * sy.TimeSpan.SECOND,
            4 * sy.TimeSpan.SECOND,
            5 * sy.TimeSpan.SECOND,
        ]
        src_data = [1.0, 2.0, 3.0, 4.0, 5.0]
        client.write(
            1 * sy.TimeSpan.SECOND,
            {
                timestamp_channel.key: idx_data,
                sensor_1.key: src_data,
            },
        )

        # Read the top-level calculated channel D
        res = client.read(
            tr=sy.TimeRange.MAX,
            channels=[calc_d.key, calc_d.index],
        )

        # Verify results
        # sensor_1 = [1, 2, 3, 4, 5]
        # calc_b = sensor_1 * 2 = [2, 4, 6, 8, 10]
        # calc_c = calc_b + 5 = [7, 9, 11, 13, 15]
        # calc_d = calc_c * 3 = [21, 27, 33, 39, 45]
        assert calc_d.key in res.channels
        assert calc_d.index in res.channels

        data_ser = res[calc_d.key]
        ts_ser = res[calc_d.index]
        assert len(ts_ser) == 5
        assert np.array_equal(
            data_ser, np.array([21.0, 27.0, 33.0, 39.0, 45.0], dtype=np.float32)
        )
        assert np.array_equal(ts_ser, np.array(idx_data, dtype=ts_ser.data_type.np))

    def test_nested_calculated_channels_diamond(self, client: sy.Synnax):
        """Should correctly handle diamond dependency pattern (E → C & D → A)"""
        # Create index and base concrete channel
        timestamp_channel = client.channels.create(
            name=random_name(),
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
        )
        sensor_1 = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=timestamp_channel.key,
                data_type=sy.DataType.FLOAT32,
            ),
        )

        # Create C: depends on sensor_1 (concrete)
        calc_c = client.channels.create(
            name=random_name(),
            expression=f"return {sensor_1.name} + 10",
        )

        # Create D: also depends on sensor_1 (concrete)
        calc_d = client.channels.create(
            name=random_name(),
            expression=f"return {sensor_1.name} * 5",
        )

        # Create E: depends on both C and D (calculated)
        calc_e = client.channels.create(
            name=random_name(),
            expression=f"return {calc_c.name} + {calc_d.name}",
        )

        # Write data
        idx_data = [
            1 * sy.TimeSpan.SECOND,
            2 * sy.TimeSpan.SECOND,
            3 * sy.TimeSpan.SECOND,
            4 * sy.TimeSpan.SECOND,
            5 * sy.TimeSpan.SECOND,
        ]
        src_data = [1.0, 2.0, 3.0, 4.0, 5.0]
        client.write(
            1 * sy.TimeSpan.SECOND,
            {
                timestamp_channel.key: idx_data,
                sensor_1.key: src_data,
            },
        )

        # Read the top-level calculated channel E
        res = client.read(
            tr=sy.TimeRange.MAX,
            channels=[calc_e.key, calc_e.index],
        )

        # Verify results
        # sensor_1 = [1, 2, 3, 4, 5]
        # calc_c = sensor_1 + 10 = [11, 12, 13, 14, 15]
        # calc_d = sensor_1 * 5 = [5, 10, 15, 20, 25]
        # calc_e = calc_c + calc_d = [16, 22, 28, 34, 40]
        assert calc_e.key in res.channels
        assert calc_e.index in res.channels

        data_ser = res[calc_e.key]
        ts_ser = res[calc_e.index]
        assert len(ts_ser) == 5
        assert np.array_equal(
            data_ser, np.array([16.0, 22.0, 28.0, 34.0, 40.0], dtype=np.float32)
        )
        assert np.array_equal(ts_ser, np.array(idx_data, dtype=ts_ser.data_type.np))

    def test_circular_dependency_detection(self, client: sy.Synnax):
        """Should detect circular dependencies at channel creation time"""
        # In Python/API, Arc validates channel references at creation time,
        # so circular dependencies are caught immediately when trying to create
        # a channel that references a non-existent channel.

        # Trying to create a channel that references a non-existent channel fails
        with pytest.raises(Exception, match="undefined symbol"):
            client.channels.create(
                name="calc_circ_invalid",
                expression="return nonexistent_channel + 1",
            )

    def test_mixed_calculated_and_concrete_channels(self, client: sy.Synnax):
        """Should correctly handle requesting both calculated and concrete channels together"""
        # Create index and base concrete channels
        timestamp_channel = client.channels.create(
            name=random_name(),
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
        )
        sensor_1 = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=timestamp_channel.key,
                data_type=sy.DataType.FLOAT32,
            ),
        )
        sensor_2 = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=timestamp_channel.key,
                data_type=sy.DataType.FLOAT32,
            ),
        )

        # Create calculated channel that depends on both sensors
        calc_mixed = client.channels.create(
            name=random_name(),
            expression=f"return {sensor_1.name} + {sensor_2.name}",
        )

        # Create nested calculated channel
        calc_mixed_nested = client.channels.create(
            name=random_name(),
            expression=f"return {calc_mixed.name} * 2",
        )

        # Write data
        idx_data = [
            1 * sy.TimeSpan.SECOND,
            2 * sy.TimeSpan.SECOND,
            3 * sy.TimeSpan.SECOND,
            4 * sy.TimeSpan.SECOND,
            5 * sy.TimeSpan.SECOND,
        ]
        src_data_1 = [1.0, 2.0, 3.0, 4.0, 5.0]
        src_data_2 = [-2.0, -3.0, -4.0, -5.0, -6.0]
        client.write(
            1 * sy.TimeSpan.SECOND,
            {
                timestamp_channel.key: idx_data,
                sensor_1.key: src_data_1,
                sensor_2.key: src_data_2,
            },
        )

        # Read both concrete and calculated channels together
        res = client.read(
            tr=sy.TimeRange.MAX,
            channels=[
                sensor_1.key,
                sensor_2.key,
                calc_mixed_nested.key,
                calc_mixed_nested.index,
            ],
        )

        # Verify concrete channels have original values
        assert sensor_1.key in res.channels
        data_ser = res[sensor_1.key]
        assert np.array_equal(data_ser, np.array(src_data_1, dtype=np.float32))

        assert sensor_2.key in res.channels
        data_ser = res[sensor_2.key]
        assert np.array_equal(data_ser, np.array(src_data_2, dtype=np.float32))

        # Verify calculated channel has correct values
        # sensor_1 = [1, 2, 3, 4, 5]
        # sensor_2 = [-2, -3, -4, -5, -6]
        # calc_mixed = sensor_1 + sensor_2 = [-1, -1, -1, -1, -1]
        # calc_mixed_nested = calc_mixed * 2 = [-2, -2, -2, -2, -2]
        assert calc_mixed_nested.key in res.channels
        data_ser = res[calc_mixed_nested.key]
        assert np.array_equal(
            data_ser, np.array([-2.0, -2.0, -2.0, -2.0, -2.0], dtype=np.float32)
        )


@pytest.mark.framer
@pytest.mark.calculations
class TestCalculationOperations:
    def test_avg_operation_accumulates_across_batches(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="avg")],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(start, [idx.key, data.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 1 * sy.TimeSpan.SECOND,
                                start + 2 * sy.TimeSpan.SECOND,
                                start + 3 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([10.0, 20.0, 30.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(20.0)
                ser = frame[calc.key]
                assert ser.alignment != 0
                first_alignment = ser.alignment
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 4 * sy.TimeSpan.SECOND,
                                start + 5 * sy.TimeSpan.SECOND,
                                start + 6 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([40.0, 50.0, 60.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(35.0)
                ser = frame[calc.key]
                assert ser.alignment == first_alignment + 1

    def test_avg_duration_reset_triggers_on_boundary(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[
                sy.channel.Operation(type="avg", duration=5 * sy.TimeSpan.SECOND)
            ],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(start, [idx.key, data.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 1 * sy.TimeSpan.SECOND,
                                start + 2 * sy.TimeSpan.SECOND,
                                start + 3 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([10.0, 20.0, 30.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(20.0)
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 6 * sy.TimeSpan.SECOND,
                                start + 7 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([40.0, 50.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(45.0)

    def test_min_operation_duration_reset(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[
                sy.channel.Operation(type="min", duration=5 * sy.TimeSpan.SECOND)
            ],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(start, [idx.key, data.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 1 * sy.TimeSpan.SECOND,
                                start + 2 * sy.TimeSpan.SECOND,
                                start + 3 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([30.0, 10.0, 20.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(10.0)
                ser = frame[calc.key]
                assert ser.alignment != 0
                first_alignment = ser.alignment
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 6 * sy.TimeSpan.SECOND,
                                start + 7 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([50.0, 40.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(40.0)
                ser = frame[calc.key]
                assert ser.alignment == first_alignment + 1

    def test_max_operation_duration_reset(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[
                sy.channel.Operation(type="max", duration=5 * sy.TimeSpan.SECOND)
            ],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(start, [idx.key, data.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 1 * sy.TimeSpan.SECOND,
                                start + 2 * sy.TimeSpan.SECOND,
                                start + 3 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([10.0, 30.0, 20.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(30.0)
                ser = frame[calc.key]
                assert ser.alignment != 0
                first_alignment = ser.alignment
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 6 * sy.TimeSpan.SECOND,
                                start + 7 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([15.0, 25.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(25.0)
                ser = frame[calc.key]
                assert ser.alignment == first_alignment + 1

    def test_avg_signal_reset_triggers_on_channel(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        reset = client.channels.create(
            name=random_name(), data_type=sy.DataType.UINT8, virtual=True
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="avg", reset_channel=reset.key)],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(start, [idx.key, data.key, reset.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 1 * sy.TimeSpan.SECOND,
                                start + 2 * sy.TimeSpan.SECOND,
                                start + 3 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([10.0, 20.0, 30.0], dtype=np.float32),
                        reset.key: np.array([0, 0, 0], dtype=np.uint8),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(20.0)
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 4 * sy.TimeSpan.SECOND,
                                start + 5 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([40.0, 50.0], dtype=np.float32),
                        reset.key: np.array([1, 0], dtype=np.uint8),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(45.0)

    def test_min_signal_reset_clears_state(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        reset = client.channels.create(
            name=random_name(), data_type=sy.DataType.UINT8, virtual=True
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="min", reset_channel=reset.key)],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(start, [idx.key, data.key, reset.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 1 * sy.TimeSpan.SECOND,
                                start + 2 * sy.TimeSpan.SECOND,
                                start + 3 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([30.0, 10.0, 20.0], dtype=np.float32),
                        reset.key: np.array([0, 0, 0], dtype=np.uint8),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(10.0)
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 4 * sy.TimeSpan.SECOND,
                                start + 5 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([50.0, 40.0], dtype=np.float32),
                        reset.key: np.array([1, 0], dtype=np.uint8),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(40.0)

    def test_max_signal_reset_clears_state(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        reset = client.channels.create(
            name=random_name(), data_type=sy.DataType.UINT8, virtual=True
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="max", reset_channel=reset.key)],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(start, [idx.key, data.key, reset.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 1 * sy.TimeSpan.SECOND,
                                start + 2 * sy.TimeSpan.SECOND,
                                start + 3 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([10.0, 30.0, 20.0], dtype=np.float32),
                        reset.key: np.array([0, 0, 0], dtype=np.uint8),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(30.0)
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 4 * sy.TimeSpan.SECOND,
                                start + 5 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([15.0, 25.0], dtype=np.float32),
                        reset.key: np.array([1, 0], dtype=np.uint8),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert len(frame[calc.key]) == 1
                assert frame[calc.key][0] == pytest.approx(25.0)

    def test_operations_with_single_sample(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        calc_min = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="min")],
        )
        calc_max = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="max")],
        )
        calc_avg = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="avg")],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(
            [calc_min.key, calc_max.key, calc_avg.key]
        ) as streamer:
            with client.open_writer(start, [idx.key, data.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [start + 1 * sy.TimeSpan.SECOND], dtype=np.int64
                        ),
                        data.key: np.array([42.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert frame[calc_min.key][0] == pytest.approx(42.0)
                assert frame[calc_max.key][0] == pytest.approx(42.0)
                assert frame[calc_avg.key][0] == pytest.approx(42.0)

    def test_avg_multiple_duration_resets(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[
                sy.channel.Operation(type="avg", duration=3 * sy.TimeSpan.SECOND)
            ],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(start, [idx.key, data.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [start + 1 * sy.TimeSpan.SECOND], dtype=np.int64
                        ),
                        data.key: np.array([10.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert frame[calc.key][0] == pytest.approx(10.0)
                writer.write(
                    {
                        idx.key: np.array(
                            [start + 4 * sy.TimeSpan.SECOND], dtype=np.int64
                        ),
                        data.key: np.array([20.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert frame[calc.key][0] == pytest.approx(20.0)
                writer.write(
                    {
                        idx.key: np.array(
                            [start + 7 * sy.TimeSpan.SECOND], dtype=np.int64
                        ),
                        data.key: np.array([30.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert frame[calc.key][0] == pytest.approx(30.0)

    def test_avg_combined_duration_and_signal_reset(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        reset = client.channels.create(
            name=random_name(), data_type=sy.DataType.UINT8, virtual=True
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[
                sy.channel.Operation(
                    type="avg",
                    duration=10 * sy.TimeSpan.SECOND,
                    reset_channel=reset.key,
                )
            ],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(start, [idx.key, data.key, reset.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 1 * sy.TimeSpan.SECOND,
                                start + 2 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([10.0, 20.0], dtype=np.float32),
                        reset.key: np.array([0, 0], dtype=np.uint8),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert frame[calc.key][0] == pytest.approx(15.0)
                writer.write(
                    {
                        idx.key: np.array(
                            [start + 3 * sy.TimeSpan.SECOND], dtype=np.int64
                        ),
                        data.key: np.array([30.0], dtype=np.float32),
                        reset.key: np.array([1], dtype=np.uint8),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert frame[calc.key][0] == pytest.approx(30.0)
                writer.write(
                    {
                        idx.key: np.array(
                            [start + 4 * sy.TimeSpan.SECOND], dtype=np.int64
                        ),
                        data.key: np.array([40.0], dtype=np.float32),
                        reset.key: np.array([0], dtype=np.uint8),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert frame[calc.key][0] == pytest.approx(35.0)

    def test_operations_with_identical_values(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        calc_min = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="min")],
        )
        calc_max = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="max")],
        )
        calc_avg = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="avg")],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(
            [calc_min.key, calc_max.key, calc_avg.key]
        ) as streamer:
            with client.open_writer(start, [idx.key, data.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 1 * sy.TimeSpan.SECOND,
                                start + 2 * sy.TimeSpan.SECOND,
                                start + 3 * sy.TimeSpan.SECOND,
                                start + 4 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([5.0, 5.0, 5.0, 5.0], dtype=np.float32),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert frame[calc_min.key][0] == pytest.approx(5.0)
                assert frame[calc_max.key][0] == pytest.approx(5.0)
                assert frame[calc_avg.key][0] == pytest.approx(5.0)

    def test_avg_signal_reset_fast_pulse(self, client: sy.Synnax):
        idx = client.channels.create(
            name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
        )
        data = client.channels.create(
            name=random_name(), data_type=sy.DataType.FLOAT32, index=idx.key
        )
        reset = client.channels.create(
            name=random_name(), data_type=sy.DataType.UINT8, virtual=True
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="avg", reset_channel=reset.key)],
        )
        start = sy.TimeStamp.now()
        with client.open_streamer(calc.key) as streamer:
            with client.open_writer(start, [idx.key, data.key, reset.key]) as writer:
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 1 * sy.TimeSpan.SECOND,
                                start + 2 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([10.0, 20.0], dtype=np.float32),
                        reset.key: np.array([0, 0], dtype=np.uint8),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert frame[calc.key][0] == pytest.approx(15.0)
                writer.write(
                    {
                        idx.key: np.array(
                            [
                                start + 3 * sy.TimeSpan.SECOND,
                                start + 4 * sy.TimeSpan.SECOND,
                                start + 5 * sy.TimeSpan.SECOND,
                            ],
                            dtype=np.int64,
                        ),
                        data.key: np.array([30.0, 40.0, 50.0], dtype=np.float32),
                        reset.key: np.array([1, 0, 1], dtype=np.uint8),
                    }
                )
                frame = streamer.read(timeout=1)
                assert frame is not None
                assert frame[calc.key][0] == pytest.approx(40.0)


@pytest.mark.framer
@pytest.mark.calculations
class TestCalculationsAcrossDomains:
    """Tests for calculated channels reading data written across multiple domains.

    These tests verify that calculations work correctly when data is written
    using multiple separate writers with auto-commit, creating distinct domains.
    Each test verifies both the calculated values AND the alignment metadata.
    """

    def test_basic_calculation_across_two_domains(self, client: sy.Synnax):
        """Should correctly calculate values across two separate write domains
        with correct alignment for each domain."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        data = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name} * 2",
        )

        # First domain: write with auto-commit, then close
        with client.open_writer(
            1 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            1 * sy.TimeSpan.SECOND,
                            2 * sy.TimeSpan.SECOND,
                            3 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    data.key: np.array([1.0, 2.0, 3.0], dtype=np.float32),
                }
            )

        # Second domain: write with auto-commit, then close
        with client.open_writer(
            6 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            6 * sy.TimeSpan.SECOND,
                            7 * sy.TimeSpan.SECOND,
                            8 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    data.key: np.array([6.0, 7.0, 8.0], dtype=np.float32),
                }
            )

        # Read calculated channel - should have data from both domains
        res = client.read(sy.TimeRange.MAX, [calc.key, calc.index])

        # Access the MultiSeries to check individual domain series
        calc_multi = res[calc.key]
        idx_multi = res[calc.index]

        # Should have 2 series (one per domain)
        assert len(calc_multi.series) == 2
        assert len(idx_multi.series) == 2

        # Verify domain 0 alignment and values
        assert calc_multi.series[0].alignment.domain_index == 0
        assert calc_multi.series[0].alignment.sample_index == 0
        assert idx_multi.series[0].alignment.domain_index == 0
        assert idx_multi.series[0].alignment.sample_index == 0
        assert np.allclose(
            calc_multi.series[0], np.array([2.0, 4.0, 6.0], dtype=np.float32)
        )

        # Verify domain 1 alignment and values
        assert calc_multi.series[1].alignment.domain_index == 1
        assert calc_multi.series[1].alignment.sample_index == 0
        assert idx_multi.series[1].alignment.domain_index == 1
        assert idx_multi.series[1].alignment.sample_index == 0
        assert np.allclose(
            calc_multi.series[1], np.array([12.0, 14.0, 16.0], dtype=np.float32)
        )

        # Verify calc and index alignments match within each domain
        assert calc_multi.series[0].alignment == idx_multi.series[0].alignment
        assert calc_multi.series[1].alignment == idx_multi.series[1].alignment

    def test_nested_calculation_2_level_across_domains(self, client: sy.Synnax):
        """Should correctly handle 2-level nested calculations across domains
        with proper alignment preservation."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        sensor = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        # B depends on sensor (concrete)
        calc_b = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {sensor.name} * 2",
        )
        # C depends on B (calculated)
        calc_c = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {calc_b.name} + 10",
        )

        # First domain
        with client.open_writer(
            1 * sy.TimeSpan.SECOND, [idx.key, sensor.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            1 * sy.TimeSpan.SECOND,
                            2 * sy.TimeSpan.SECOND,
                            3 * sy.TimeSpan.SECOND,
                            4 * sy.TimeSpan.SECOND,
                            5 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    sensor.key: np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32),
                }
            )

        # Second domain
        with client.open_writer(
            6 * sy.TimeSpan.SECOND, [idx.key, sensor.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            6 * sy.TimeSpan.SECOND,
                            7 * sy.TimeSpan.SECOND,
                            8 * sy.TimeSpan.SECOND,
                            9 * sy.TimeSpan.SECOND,
                            10 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    sensor.key: np.array([6.0, 7.0, 8.0, 9.0, 10.0], dtype=np.float32),
                }
            )

        # Read the top-level calculated channel C
        res = client.read(sy.TimeRange.MAX, [calc_c.key, calc_c.index])
        calc_multi = res[calc_c.key]
        idx_multi = res[calc_c.index]

        # Should have 2 series (one per domain)
        assert len(calc_multi.series) == 2
        assert len(idx_multi.series) == 2

        # Verify domain 0 alignment and values
        # sensor = [1,2,3,4,5], calc_b = [2,4,6,8,10], calc_c = [12,14,16,18,20]
        assert calc_multi.series[0].alignment.domain_index == 0
        assert calc_multi.series[0].alignment.sample_index == 0
        assert np.allclose(
            calc_multi.series[0],
            np.array([12.0, 14.0, 16.0, 18.0, 20.0], dtype=np.float32),
        )

        # Verify domain 1 alignment and values
        # sensor = [6,7,8,9,10], calc_b = [12,14,16,18,20], calc_c = [22,24,26,28,30]
        assert calc_multi.series[1].alignment.domain_index == 1
        assert calc_multi.series[1].alignment.sample_index == 0
        assert np.allclose(
            calc_multi.series[1],
            np.array([22.0, 24.0, 26.0, 28.0, 30.0], dtype=np.float32),
        )

        # Verify calc and index alignments match
        assert calc_multi.series[0].alignment == idx_multi.series[0].alignment
        assert calc_multi.series[1].alignment == idx_multi.series[1].alignment

    def test_nested_calculation_3_level_across_domains(self, client: sy.Synnax):
        """Should correctly handle 3-level nested calculations across domains
        with proper alignment preservation."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        sensor = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        # B depends on sensor (concrete)
        calc_b = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {sensor.name} * 2",
        )
        # C depends on B (calculated)
        calc_c = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {calc_b.name} + 5",
        )
        # D depends on C (calculated)
        calc_d = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {calc_c.name} * 3",
        )

        # First domain
        with client.open_writer(
            1 * sy.TimeSpan.SECOND, [idx.key, sensor.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            1 * sy.TimeSpan.SECOND,
                            2 * sy.TimeSpan.SECOND,
                            3 * sy.TimeSpan.SECOND,
                            4 * sy.TimeSpan.SECOND,
                            5 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    sensor.key: np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32),
                }
            )

        # Second domain
        with client.open_writer(
            6 * sy.TimeSpan.SECOND, [idx.key, sensor.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            6 * sy.TimeSpan.SECOND,
                            7 * sy.TimeSpan.SECOND,
                            8 * sy.TimeSpan.SECOND,
                            9 * sy.TimeSpan.SECOND,
                            10 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    sensor.key: np.array([6.0, 7.0, 8.0, 9.0, 10.0], dtype=np.float32),
                }
            )

        # Read the top-level calculated channel D
        res = client.read(sy.TimeRange.MAX, [calc_d.key, calc_d.index])
        calc_multi = res[calc_d.key]
        idx_multi = res[calc_d.index]

        # Should have 2 series (one per domain)
        assert len(calc_multi.series) == 2
        assert len(idx_multi.series) == 2

        # Verify domain 0 alignment and values
        # sensor = [1,2,3,4,5]
        # calc_b = sensor * 2 = [2,4,6,8,10]
        # calc_c = calc_b + 5 = [7,9,11,13,15]
        # calc_d = calc_c * 3 = [21,27,33,39,45]
        assert calc_multi.series[0].alignment.domain_index == 0
        assert calc_multi.series[0].alignment.sample_index == 0
        assert np.allclose(
            calc_multi.series[0],
            np.array([21.0, 27.0, 33.0, 39.0, 45.0], dtype=np.float32),
        )

        # Verify domain 1 alignment and values
        # sensor = [6,7,8,9,10]
        # calc_b = sensor * 2 = [12,14,16,18,20]
        # calc_c = calc_b + 5 = [17,19,21,23,25]
        # calc_d = calc_c * 3 = [51,57,63,69,75]
        assert calc_multi.series[1].alignment.domain_index == 1
        assert calc_multi.series[1].alignment.sample_index == 0
        assert np.allclose(
            calc_multi.series[1],
            np.array([51.0, 57.0, 63.0, 69.0, 75.0], dtype=np.float32),
        )

        # Verify calc and index alignments match
        assert calc_multi.series[0].alignment == idx_multi.series[0].alignment
        assert calc_multi.series[1].alignment == idx_multi.series[1].alignment

    def test_diamond_dependency_across_domains(self, client: sy.Synnax):
        """Should correctly handle diamond dependency pattern across domains
        with proper alignment preservation.

        Diamond pattern:
        E depends on C and D
        C depends on sensor (A)
        D depends on sensor (A)
        """
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        sensor = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        # C depends on sensor
        calc_c = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {sensor.name} + 10",
        )
        # D also depends on sensor
        calc_d = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {sensor.name} * 5",
        )
        # E depends on both C and D
        calc_e = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {calc_c.name} + {calc_d.name}",
        )

        # First domain
        with client.open_writer(
            1 * sy.TimeSpan.SECOND, [idx.key, sensor.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            1 * sy.TimeSpan.SECOND,
                            2 * sy.TimeSpan.SECOND,
                            3 * sy.TimeSpan.SECOND,
                            4 * sy.TimeSpan.SECOND,
                            5 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    sensor.key: np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32),
                }
            )

        # Second domain
        with client.open_writer(
            6 * sy.TimeSpan.SECOND, [idx.key, sensor.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            6 * sy.TimeSpan.SECOND,
                            7 * sy.TimeSpan.SECOND,
                            8 * sy.TimeSpan.SECOND,
                            9 * sy.TimeSpan.SECOND,
                            10 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    sensor.key: np.array([6.0, 7.0, 8.0, 9.0, 10.0], dtype=np.float32),
                }
            )

        # Read the top-level calculated channel E
        res = client.read(sy.TimeRange.MAX, [calc_e.key, calc_e.index])
        calc_multi = res[calc_e.key]
        idx_multi = res[calc_e.index]

        # Should have 2 series (one per domain)
        assert len(calc_multi.series) == 2
        assert len(idx_multi.series) == 2

        # Verify domain 0 alignment and values
        # sensor = [1,2,3,4,5]
        # calc_c = sensor + 10 = [11,12,13,14,15]
        # calc_d = sensor * 5 = [5,10,15,20,25]
        # calc_e = calc_c + calc_d = [16,22,28,34,40]
        assert calc_multi.series[0].alignment.domain_index == 0
        assert calc_multi.series[0].alignment.sample_index == 0
        assert np.allclose(
            calc_multi.series[0],
            np.array([16.0, 22.0, 28.0, 34.0, 40.0], dtype=np.float32),
        )

        # Verify domain 1 alignment and values
        # sensor = [6,7,8,9,10]
        # calc_c = sensor + 10 = [16,17,18,19,20]
        # calc_d = sensor * 5 = [30,35,40,45,50]
        # calc_e = calc_c + calc_d = [46,52,58,64,70]
        # Note: Diamond pattern may cause alignment increment in some implementations
        assert calc_multi.series[1].alignment.sample_index == 0
        assert np.allclose(
            calc_multi.series[1],
            np.array([46.0, 52.0, 58.0, 64.0, 70.0], dtype=np.float32),
        )

        # Verify calc and index alignments match within each domain
        assert calc_multi.series[0].alignment == idx_multi.series[0].alignment
        assert calc_multi.series[1].alignment == idx_multi.series[1].alignment

    def test_mixed_calculated_and_concrete_across_domains(self, client: sy.Synnax):
        """Should correctly handle requesting both calculated and concrete channels
        with proper alignment for each."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        sensor_1 = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        sensor_2 = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        # Calculated channel that depends on both sensors
        calc_mixed = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {sensor_1.name} + {sensor_2.name}",
        )
        # Nested calculated channel
        calc_nested = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {calc_mixed.name} * 2",
        )

        # First domain
        with client.open_writer(
            1 * sy.TimeSpan.SECOND,
            [idx.key, sensor_1.key, sensor_2.key],
            enable_auto_commit=True,
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            1 * sy.TimeSpan.SECOND,
                            2 * sy.TimeSpan.SECOND,
                            3 * sy.TimeSpan.SECOND,
                            4 * sy.TimeSpan.SECOND,
                            5 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    sensor_1.key: np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32),
                    sensor_2.key: np.array(
                        [-2.0, -3.0, -4.0, -5.0, -6.0], dtype=np.float32
                    ),
                }
            )

        # Second domain
        with client.open_writer(
            6 * sy.TimeSpan.SECOND,
            [idx.key, sensor_1.key, sensor_2.key],
            enable_auto_commit=True,
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            6 * sy.TimeSpan.SECOND,
                            7 * sy.TimeSpan.SECOND,
                            8 * sy.TimeSpan.SECOND,
                            9 * sy.TimeSpan.SECOND,
                            10 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    sensor_1.key: np.array(
                        [6.0, 7.0, 8.0, 9.0, 10.0], dtype=np.float32
                    ),
                    sensor_2.key: np.array(
                        [-3.0, -4.0, -5.0, -6.0, -7.0], dtype=np.float32
                    ),
                }
            )

        # Read both concrete and calculated channels together
        res = client.read(
            sy.TimeRange.MAX,
            [sensor_1.key, sensor_2.key, calc_nested.key, calc_nested.index],
        )

        s1_multi = res[sensor_1.key]
        s2_multi = res[sensor_2.key]
        calc_multi = res[calc_nested.key]
        idx_multi = res[calc_nested.index]

        # All should have 2 series (one per domain)
        assert len(s1_multi.series) == 2
        assert len(s2_multi.series) == 2
        assert len(calc_multi.series) == 2
        assert len(idx_multi.series) == 2

        # Verify domain 0 alignments for concrete channels
        assert s1_multi.series[0].alignment.domain_index == 0
        assert s2_multi.series[0].alignment.domain_index == 0
        # Calculated channels have summed alignment: calc_mixed uses 2 inputs,
        # calc_nested uses calc_mixed. For domain 0: (0+0)*2 = 0
        assert calc_multi.series[0].alignment.domain_index == 0
        assert idx_multi.series[0].alignment.domain_index == 0

        # Verify domain 1 alignments for concrete channels
        assert s1_multi.series[1].alignment.domain_index == 1
        assert s2_multi.series[1].alignment.domain_index == 1
        # For domain 1: calc_mixed = (1+1) = 2, calc_nested uses calc_mixed = 2
        assert calc_multi.series[1].alignment.domain_index == 2
        assert idx_multi.series[1].alignment.domain_index == 2

        # Verify concrete channel values
        assert np.allclose(
            s1_multi.series[0], np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32)
        )
        assert np.allclose(
            s1_multi.series[1], np.array([6.0, 7.0, 8.0, 9.0, 10.0], dtype=np.float32)
        )
        assert np.allclose(
            s2_multi.series[0],
            np.array([-2.0, -3.0, -4.0, -5.0, -6.0], dtype=np.float32),
        )
        assert np.allclose(
            s2_multi.series[1],
            np.array([-3.0, -4.0, -5.0, -6.0, -7.0], dtype=np.float32),
        )

        # Verify calculated channel values
        # sensor_1 = [1,2,3,4,5] and [6,7,8,9,10]
        # sensor_2 = [-2,-3,-4,-5,-6] and [-3,-4,-5,-6,-7]
        # calc_mixed = sensor_1 + sensor_2 = [-1,-1,-1,-1,-1] and [3,3,3,3,3]
        # calc_nested = calc_mixed * 2 = [-2,-2,-2,-2,-2] and [6,6,6,6,6]
        assert np.allclose(
            calc_multi.series[0],
            np.array([-2.0, -2.0, -2.0, -2.0, -2.0], dtype=np.float32),
        )
        assert np.allclose(
            calc_multi.series[1],
            np.array([6.0, 6.0, 6.0, 6.0, 6.0], dtype=np.float32),
        )

    def test_three_domains_calculation(self, client: sy.Synnax):
        """Should correctly calculate values across three separate write domains
        with correct alignment for each."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        data = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name} * 2",
        )

        # First domain
        with client.open_writer(
            1 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [1 * sy.TimeSpan.SECOND, 2 * sy.TimeSpan.SECOND],
                        dtype=np.int64,
                    ),
                    data.key: np.array([1.0, 2.0], dtype=np.float32),
                }
            )

        # Second domain
        with client.open_writer(
            5 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [5 * sy.TimeSpan.SECOND, 6 * sy.TimeSpan.SECOND],
                        dtype=np.int64,
                    ),
                    data.key: np.array([5.0, 6.0], dtype=np.float32),
                }
            )

        # Third domain
        with client.open_writer(
            10 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [10 * sy.TimeSpan.SECOND, 11 * sy.TimeSpan.SECOND],
                        dtype=np.int64,
                    ),
                    data.key: np.array([10.0, 11.0], dtype=np.float32),
                }
            )

        # Read calculated channel
        res = client.read(sy.TimeRange.MAX, [calc.key, calc.index])
        calc_multi = res[calc.key]
        idx_multi = res[calc.index]

        # Should have 3 series (one per domain)
        assert len(calc_multi.series) == 3
        assert len(idx_multi.series) == 3

        # Verify domain 0
        assert calc_multi.series[0].alignment.domain_index == 0
        assert calc_multi.series[0].alignment.sample_index == 0
        assert np.allclose(calc_multi.series[0], np.array([2.0, 4.0], dtype=np.float32))

        # Verify domain 1
        assert calc_multi.series[1].alignment.domain_index == 1
        assert calc_multi.series[1].alignment.sample_index == 0
        assert np.allclose(
            calc_multi.series[1], np.array([10.0, 12.0], dtype=np.float32)
        )

        # Verify domain 2
        assert calc_multi.series[2].alignment.domain_index == 2
        assert calc_multi.series[2].alignment.sample_index == 0
        assert np.allclose(
            calc_multi.series[2], np.array([20.0, 22.0], dtype=np.float32)
        )

        # Verify all alignments match between calc and index
        for i in range(3):
            assert calc_multi.series[i].alignment == idx_multi.series[i].alignment

    def test_calculation_with_gap_between_domains(self, client: sy.Synnax):
        """Should correctly handle calculations when there's a large gap between domains
        while preserving proper alignment."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        data = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name} + 100",
        )

        # First domain at t=1s
        with client.open_writer(
            1 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            1 * sy.TimeSpan.SECOND,
                            2 * sy.TimeSpan.SECOND,
                            3 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    data.key: np.array([1.0, 2.0, 3.0], dtype=np.float32),
                }
            )

        # Second domain at t=1000s (large gap)
        with client.open_writer(
            1000 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            1000 * sy.TimeSpan.SECOND,
                            1001 * sy.TimeSpan.SECOND,
                            1002 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    data.key: np.array([1000.0, 1001.0, 1002.0], dtype=np.float32),
                }
            )

        # Read calculated channel
        res = client.read(sy.TimeRange.MAX, [calc.key, calc.index])
        calc_multi = res[calc.key]
        idx_multi = res[calc.index]

        # Should have 2 series despite the large gap
        assert len(calc_multi.series) == 2
        assert len(idx_multi.series) == 2

        # Verify domain 0
        assert calc_multi.series[0].alignment.domain_index == 0
        assert np.allclose(
            calc_multi.series[0], np.array([101.0, 102.0, 103.0], dtype=np.float32)
        )

        # Verify domain 1
        assert calc_multi.series[1].alignment.domain_index == 1
        assert np.allclose(
            calc_multi.series[1],
            np.array([1100.0, 1101.0, 1102.0], dtype=np.float32),
        )

    def test_multiple_calculations_same_source_across_domains(self, client: sy.Synnax):
        """Should correctly handle multiple calculations depending on the same source
        with consistent alignments across all calculated channels."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        sensor = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        # Multiple calculations on the same source
        calc_double = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {sensor.name} * 2",
        )
        calc_square = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {sensor.name} * {sensor.name}",
        )
        calc_plus_ten = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {sensor.name} + 10",
        )

        # First domain
        with client.open_writer(
            1 * sy.TimeSpan.SECOND, [idx.key, sensor.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [1 * sy.TimeSpan.SECOND, 2 * sy.TimeSpan.SECOND],
                        dtype=np.int64,
                    ),
                    sensor.key: np.array([2.0, 3.0], dtype=np.float32),
                }
            )

        # Second domain
        with client.open_writer(
            5 * sy.TimeSpan.SECOND, [idx.key, sensor.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [5 * sy.TimeSpan.SECOND, 6 * sy.TimeSpan.SECOND],
                        dtype=np.int64,
                    ),
                    sensor.key: np.array([4.0, 5.0], dtype=np.float32),
                }
            )

        # Read all calculated channels
        res = client.read(
            sy.TimeRange.MAX, [calc_double.key, calc_square.key, calc_plus_ten.key]
        )

        double_multi = res[calc_double.key]
        square_multi = res[calc_square.key]
        plus_ten_multi = res[calc_plus_ten.key]

        # All should have 2 series
        assert len(double_multi.series) == 2
        assert len(square_multi.series) == 2
        assert len(plus_ten_multi.series) == 2

        # Verify all have consistent domain alignments
        for i in range(2):
            assert double_multi.series[i].alignment.domain_index == i
            assert square_multi.series[i].alignment.domain_index == i
            assert plus_ten_multi.series[i].alignment.domain_index == i

        # Verify calc_double: sensor * 2
        assert np.allclose(
            double_multi.series[0], np.array([4.0, 6.0], dtype=np.float32)
        )
        assert np.allclose(
            double_multi.series[1], np.array([8.0, 10.0], dtype=np.float32)
        )

        # Verify calc_square: sensor * sensor
        assert np.allclose(
            square_multi.series[0], np.array([4.0, 9.0], dtype=np.float32)
        )
        assert np.allclose(
            square_multi.series[1], np.array([16.0, 25.0], dtype=np.float32)
        )

        # Verify calc_plus_ten: sensor + 10
        assert np.allclose(
            plus_ten_multi.series[0], np.array([12.0, 13.0], dtype=np.float32)
        )
        assert np.allclose(
            plus_ten_multi.series[1], np.array([14.0, 15.0], dtype=np.float32)
        )

    def test_calculation_with_conditional_across_domains(self, client: sy.Synnax):
        """Should correctly handle conditional calculations across domains
        with proper alignment."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        data = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"""
            if ({data.name} > 5) {{
                return 1
            }} else {{
                return 0
            }}
            """,
        )

        # First domain: values below threshold
        with client.open_writer(
            1 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [1 * sy.TimeSpan.SECOND, 2 * sy.TimeSpan.SECOND],
                        dtype=np.int64,
                    ),
                    data.key: np.array([3.0, 4.0], dtype=np.float32),
                }
            )

        # Second domain: values above threshold
        with client.open_writer(
            5 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [5 * sy.TimeSpan.SECOND, 6 * sy.TimeSpan.SECOND],
                        dtype=np.int64,
                    ),
                    data.key: np.array([6.0, 7.0], dtype=np.float32),
                }
            )

        # Third domain: mixed values
        with client.open_writer(
            10 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [10 * sy.TimeSpan.SECOND, 11 * sy.TimeSpan.SECOND],
                        dtype=np.int64,
                    ),
                    data.key: np.array([2.0, 10.0], dtype=np.float32),
                }
            )

        # Read calculated channel
        res = client.read(sy.TimeRange.MAX, [calc.key, calc.index])
        calc_multi = res[calc.key]
        idx_multi = res[calc.index]

        # Should have 3 series
        assert len(calc_multi.series) == 3
        assert len(idx_multi.series) == 3

        # Verify alignments
        assert calc_multi.series[0].alignment.domain_index == 0
        assert calc_multi.series[1].alignment.domain_index == 1
        assert calc_multi.series[2].alignment.domain_index == 2

        # Domain 0: [3, 4] -> [0, 0]
        assert np.allclose(calc_multi.series[0], np.array([0.0, 0.0], dtype=np.float32))

        # Domain 1: [6, 7] -> [1, 1]
        assert np.allclose(calc_multi.series[1], np.array([1.0, 1.0], dtype=np.float32))

        # Domain 2: [2, 10] -> [0, 1]
        assert np.allclose(calc_multi.series[2], np.array([0.0, 1.0], dtype=np.float32))

        # Verify calc and index alignments match
        for i in range(3):
            assert calc_multi.series[i].alignment == idx_multi.series[i].alignment

    def test_two_source_channels_same_writes_across_domains(self, client: sy.Synnax):
        """Should handle calculations with two source channels written together
        across domains with proper alignment - reading ONLY calculated channels."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        sensor_a = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        sensor_b = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {sensor_a.name} + {sensor_b.name}",
        )

        # First domain
        with client.open_writer(
            1 * sy.TimeSpan.SECOND,
            [idx.key, sensor_a.key, sensor_b.key],
            enable_auto_commit=True,
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            1 * sy.TimeSpan.SECOND,
                            2 * sy.TimeSpan.SECOND,
                            3 * sy.TimeSpan.SECOND,
                            4 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    sensor_a.key: np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float32),
                    sensor_b.key: np.array([10.0, 10.0, 10.0, 10.0], dtype=np.float32),
                }
            )

        # Second domain
        with client.open_writer(
            10 * sy.TimeSpan.SECOND,
            [idx.key, sensor_a.key, sensor_b.key],
            enable_auto_commit=True,
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            10 * sy.TimeSpan.SECOND,
                            11 * sy.TimeSpan.SECOND,
                            12 * sy.TimeSpan.SECOND,
                            13 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    sensor_a.key: np.array([10.0, 11.0, 12.0, 13.0], dtype=np.float32),
                    sensor_b.key: np.array([20.0, 20.0, 20.0, 20.0], dtype=np.float32),
                }
            )

        res = client.read(sy.TimeRange.MAX, [calc.key, calc.index])
        calc_multi = res[calc.key]
        idx_multi = res[calc.index]

        assert len(calc_multi.series) == 2
        assert len(idx_multi.series) == 2

        # Alignments are summed: (0,0)+(0,0)=(0,0), (1,0)+(1,0)=(2,0)
        assert calc_multi.series[0].alignment.domain_index == 0
        assert calc_multi.series[0].alignment.sample_index == 0
        assert calc_multi.series[1].alignment.domain_index == 2
        assert calc_multi.series[1].alignment.sample_index == 0

        # Domain 0: [1,2,3,4] + [10,10,10,10] = [11,12,13,14]
        assert np.allclose(
            calc_multi.series[0], np.array([11.0, 12.0, 13.0, 14.0], dtype=np.float32)
        )
        # Domain 1: [10,11,12,13] + [20,20,20,20] = [30,31,32,33]
        assert np.allclose(
            calc_multi.series[1], np.array([30.0, 31.0, 32.0, 33.0], dtype=np.float32)
        )

        assert calc_multi.series[0].alignment == idx_multi.series[0].alignment
        assert calc_multi.series[1].alignment == idx_multi.series[1].alignment


class TestStatOperationsAlignment:
    """Tests that statistical operations correctly propagate alignment metadata across domains."""

    def test_avg_operation_alignment_across_domains(self, client: sy.Synnax):
        """Should correctly calculate avg with proper alignment per domain."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        data = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="avg")],
        )

        # First domain
        with client.open_writer(
            1 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            1 * sy.TimeSpan.SECOND,
                            2 * sy.TimeSpan.SECOND,
                            3 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    data.key: np.array([10.0, 20.0, 30.0], dtype=np.float32),
                }
            )

        # Second domain
        with client.open_writer(
            6 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            6 * sy.TimeSpan.SECOND,
                            7 * sy.TimeSpan.SECOND,
                            8 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    data.key: np.array([40.0, 50.0, 60.0], dtype=np.float32),
                }
            )

        # Read calculated channel
        res = client.read(sy.TimeRange.MAX, [calc.key, calc.index])
        calc_multi = res[calc.key]
        idx_multi = res[calc.index]

        # Should have 2 series (one per domain)
        assert len(calc_multi.series) == 2
        assert len(idx_multi.series) == 2

        # Verify domain 0 alignment
        assert calc_multi.series[0].alignment.domain_index == 0
        assert calc_multi.series[0].alignment.sample_index == 0
        assert idx_multi.series[0].alignment.domain_index == 0

        # Verify domain 1 alignment
        assert calc_multi.series[1].alignment.domain_index == 1
        assert calc_multi.series[1].alignment.sample_index == 0
        assert idx_multi.series[1].alignment.domain_index == 1

        assert calc_multi.series[0][0] == pytest.approx(20.0)
        assert calc_multi.series[1][0] == pytest.approx(35.0)

        assert calc_multi.series[0].alignment == idx_multi.series[0].alignment
        assert calc_multi.series[1].alignment == idx_multi.series[1].alignment

    def test_min_operation_alignment_across_domains(self, client: sy.Synnax):
        """Should correctly calculate min with proper alignment per domain."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        data = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="min")],
        )

        # First domain
        with client.open_writer(
            1 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            1 * sy.TimeSpan.SECOND,
                            2 * sy.TimeSpan.SECOND,
                            3 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    data.key: np.array([30.0, 10.0, 20.0], dtype=np.float32),
                }
            )

        # Second domain
        with client.open_writer(
            6 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            6 * sy.TimeSpan.SECOND,
                            7 * sy.TimeSpan.SECOND,
                            8 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    data.key: np.array([60.0, 40.0, 50.0], dtype=np.float32),
                }
            )

        # Read calculated channel
        res = client.read(sy.TimeRange.MAX, [calc.key, calc.index])
        calc_multi = res[calc.key]
        idx_multi = res[calc.index]

        # Should have 2 series (one per domain)
        assert len(calc_multi.series) == 2
        assert len(idx_multi.series) == 2

        # Verify domain 0 alignment
        assert calc_multi.series[0].alignment.domain_index == 0
        assert calc_multi.series[0].alignment.sample_index == 0

        # Verify domain 1 alignment
        assert calc_multi.series[1].alignment.domain_index == 1
        assert calc_multi.series[1].alignment.sample_index == 0

        assert calc_multi.series[0][0] == pytest.approx(10.0)
        assert calc_multi.series[1][0] == pytest.approx(10.0)

        # Verify calc and index alignments match
        assert calc_multi.series[0].alignment == idx_multi.series[0].alignment
        assert calc_multi.series[1].alignment == idx_multi.series[1].alignment

    def test_max_operation_alignment_across_domains(self, client: sy.Synnax):
        """Should correctly calculate max with proper alignment per domain."""
        idx = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.TIMESTAMP,
            is_index=True,
        )
        data = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=idx.key,
        )
        calc = client.channels.create(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            expression=f"return {data.name}",
            operations=[sy.channel.Operation(type="max")],
        )

        # First domain
        with client.open_writer(
            1 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            1 * sy.TimeSpan.SECOND,
                            2 * sy.TimeSpan.SECOND,
                            3 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    data.key: np.array([10.0, 30.0, 20.0], dtype=np.float32),
                }
            )

        # Second domain
        with client.open_writer(
            6 * sy.TimeSpan.SECOND, [idx.key, data.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    idx.key: np.array(
                        [
                            6 * sy.TimeSpan.SECOND,
                            7 * sy.TimeSpan.SECOND,
                            8 * sy.TimeSpan.SECOND,
                        ],
                        dtype=np.int64,
                    ),
                    data.key: np.array([40.0, 60.0, 50.0], dtype=np.float32),
                }
            )

        # Read calculated channel
        res = client.read(sy.TimeRange.MAX, [calc.key, calc.index])
        calc_multi = res[calc.key]
        idx_multi = res[calc.index]

        # Should have 2 series (one per domain)
        assert len(calc_multi.series) == 2
        assert len(idx_multi.series) == 2

        # Verify domain 0 alignment
        assert calc_multi.series[0].alignment.domain_index == 0
        assert calc_multi.series[0].alignment.sample_index == 0

        # Verify domain 1 alignment
        assert calc_multi.series[1].alignment.domain_index == 1
        assert calc_multi.series[1].alignment.sample_index == 0

        assert calc_multi.series[0][0] == pytest.approx(30.0)
        assert calc_multi.series[1][0] == pytest.approx(60.0)

        # Verify calc and index alignments match
        assert calc_multi.series[0].alignment == idx_multi.series[0].alignment
        assert calc_multi.series[1].alignment == idx_multi.series[1].alignment

    @pytest.mark.focus
    def test_literal_minus_f32_should_succeed(self, client: sy.Synnax):
        """Regression: literal on left side of f32 channel should coerce to f32.

        Previously, `1000 - f32_channel` would create successfully but fail on
        read because the analyzer inferred the return type from the literal (i64)
        instead of the channel (f32). Now the literal correctly coerces.
        """
        ts = client.channels.create(
            name=random_name(),
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
        )
        src = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=ts.key,
                data_type=sy.DataType.FLOAT32,
            ),
        )
        calc = client.channels.create(
            name=random_name(),
            expression=f"return 1000 - {src.name}",
        )
        client.write(
            1 * sy.TimeSpan.SECOND,
            {
                ts.key: [1 * sy.TimeSpan.SECOND, 2 * sy.TimeSpan.SECOND],
                src.key: [10.0, 20.0],
            },
        )
        res = client.read(
            tr=sy.TimeRange.MAX,
            channels=[calc.key, calc.index],
        )
        assert len(res[calc.key]) == 2

    @pytest.mark.focus
    def test_float_literal_div_f32_should_succeed(self, client: sy.Synnax):
        """Regression: float literal on left side of f32 channel should coerce.

        Previously, `1000.0 / f32_channel` would create successfully but fail
        on read due to return type mismatch. Now the literal coerces to f32.
        """
        ts = client.channels.create(
            name=random_name(),
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
        )
        src = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=ts.key,
                data_type=sy.DataType.FLOAT32,
            ),
        )
        calc = client.channels.create(
            name=random_name(),
            expression=f"return 1000.0 / {src.name}",
        )
        client.write(
            1 * sy.TimeSpan.SECOND,
            {
                ts.key: [1 * sy.TimeSpan.SECOND, 2 * sy.TimeSpan.SECOND],
                src.key: [10.0, 20.0],
            },
        )
        res = client.read(
            tr=sy.TimeRange.MAX,
            channels=[calc.key, calc.index],
        )
        assert len(res[calc.key]) == 2

    @pytest.mark.focus
    def test_mixed_f32_f64_with_literal_should_reject_at_creation(
        self, client: sy.Synnax
    ):
        """Regression: mixed f32/f64 channels with literal should be caught at creation.

        Previously, `1000.0 - f32_ch + f64_ch` was accepted at creation because
        the literal's FloatConstraint individually accepted both f32 and f64.
        Now the analyzer correctly rejects the f32/f64 mismatch at creation time.
        """
        ts = client.channels.create(
            name=random_name(),
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
        )
        f32_ch = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=ts.key,
                data_type=sy.DataType.FLOAT32,
            ),
        )
        f64_ch = client.channels.create(
            sy.Channel(
                name=random_name(),
                index=ts.key,
                data_type=sy.DataType.FLOAT64,
            ),
        )
        with pytest.raises(Exception, match="type mismatch"):
            client.channels.create(
                name=random_name(),
                expression=f"return 1000.0 - {f32_ch.name} + {f64_ch.name}",
            )


@pytest.mark.framer
@pytest.mark.calculations
class TestCalcChannelStress:
    """Stress test: deep calc chains, multi-domain writes, writer/streamer lifecycle
    chaos, cross-index references, and full numerical verification."""

    @pytest.mark.focus
    def test_deep_chain_multi_domain_writer_streamer_chaos(self, client: sy.Synnax):
        S = sy.TimeSpan.SECOND

        # ── TOPOLOGY ──────────────────────────────────────────────────────
        # Two independent index channels to test cross-index behavior
        ts_a = client.channels.create(
            name=random_name(), is_index=True, data_type=sy.DataType.TIMESTAMP
        )
        ts_b = client.channels.create(
            name=random_name(), is_index=True, data_type=sy.DataType.TIMESTAMP
        )

        # Source channels on index A (f32)
        alpha = client.channels.create(
            sy.Channel(
                name=random_name(), index=ts_a.key, data_type=sy.DataType.FLOAT32
            )
        )
        beta = client.channels.create(
            sy.Channel(
                name=random_name(), index=ts_a.key, data_type=sy.DataType.FLOAT32
            )
        )

        # Source channel on index B (f32) - separate timeline
        gamma = client.channels.create(
            sy.Channel(
                name=random_name(), index=ts_b.key, data_type=sy.DataType.FLOAT32
            )
        )

        # ── CALC CHAIN (4 levels deep) ────────────────────────────────────
        # Level 1: direct operations on source channels
        c_offset = client.channels.create(
            name=random_name(),
            expression=f"return {alpha.name} + 10",
        )
        c_lit_left = client.channels.create(
            name=random_name(),
            expression=f"return 1000 - {alpha.name}",
        )
        c_scaled = client.channels.create(
            name=random_name(),
            expression=f"return 2.5 * {alpha.name}",
        )
        c_product = client.channels.create(
            name=random_name(),
            expression=f"return {alpha.name} * {beta.name}",
        )
        c_power = client.channels.create(
            name=random_name(),
            expression=f"return {alpha.name} ^ 2",
        )
        c_temp_conv = client.channels.create(
            name=random_name(),
            expression=f"return ({alpha.name} - 32) * 5 / 9",
        )
        c_inverse = client.channels.create(
            name=random_name(),
            expression=f"return 10000.0 / {alpha.name}",
        )

        # Calc on the separate index B
        c_gamma_offset = client.channels.create(
            name=random_name(),
            expression=f"return {gamma.name} * 3 + 7",
        )

        # Level 2: calcs referencing calcs
        c_chain2 = client.channels.create(
            name=random_name(),
            expression=f"return {c_offset.name} * 3",
        )
        c_combo2 = client.channels.create(
            name=random_name(),
            expression=f"return {c_scaled.name} + {c_product.name}",
        )

        # Level 3: deeper chain
        c_deep3 = client.channels.create(
            name=random_name(),
            expression=f"return {c_chain2.name} - {c_lit_left.name}",
        )

        # Level 4: deepest - references L3 and L1
        c_mega4 = client.channels.create(
            name=random_name(),
            expression=f"return ({c_deep3.name} + {c_inverse.name}) / 10",
        )

        # Calc referencing the L2 gamma calc (on index B)
        c_gamma_chain = client.channels.create(
            name=random_name(),
            expression=f"return {c_gamma_offset.name} ^ 2",
        )

        # ── EXPECTED VALUE FUNCTIONS ──────────────────────────────────────
        def expect_a(a_vals, b_vals):
            """Return dict of calc_key -> expected numpy array for index A calcs."""
            a = np.array(a_vals, dtype=np.float32)
            b = np.array(b_vals, dtype=np.float32)
            return {
                c_offset.key: a + 10,
                c_lit_left.key: 1000 - a,
                c_scaled.key: 2.5 * a,
                c_product.key: a * b,
                c_power.key: a**2,
                c_temp_conv.key: (a - 32) * 5 / 9,
                c_inverse.key: 10000.0 / a,
                c_chain2.key: (a + 10) * 3,
                c_combo2.key: 2.5 * a + a * b,
                c_deep3.key: (a + 10) * 3 - (1000 - a),
                c_mega4.key: (((a + 10) * 3 - (1000 - a)) + 10000.0 / a) / 10,
            }

        def expect_b(g_vals):
            """Return dict of calc_key -> expected numpy array for index B calcs."""
            g = np.array(g_vals, dtype=np.float32)
            gamma_off = g * 3 + 7
            return {
                c_gamma_offset.key: gamma_off,
                c_gamma_chain.key: gamma_off**2,
            }

        def verify_read(res, expected, label=""):
            for key, exp in expected.items():
                actual = np.array(res[key], dtype=np.float32)
                np.testing.assert_allclose(
                    actual,
                    np.array(exp, dtype=np.float32),
                    rtol=1e-4,
                    err_msg=f"{label} key={key}",
                )

        def verify_stream_frame(frame, expected, label=""):
            for key, exp in expected.items():
                if key not in frame.channels:
                    continue
                actual = np.array(frame[key], dtype=np.float32)
                np.testing.assert_allclose(
                    actual,
                    np.array(exp, dtype=np.float32),
                    rtol=1e-4,
                    err_msg=f"stream {label} key={key}",
                )

        all_a_calc_keys = [
            c_offset.key,
            c_lit_left.key,
            c_scaled.key,
            c_product.key,
            c_power.key,
            c_temp_conv.key,
            c_inverse.key,
            c_chain2.key,
            c_combo2.key,
            c_deep3.key,
            c_mega4.key,
        ]
        all_b_calc_keys = [c_gamma_offset.key, c_gamma_chain.key]

        # ── PHASE 1: Multi-domain writes, full read verification ──────────
        # Domain 1: 3 samples on index A
        a1, b1 = [10.0, 20.0, 50.0], [2.0, 4.0, 1.0]
        with client.open_writer(
            1 * S, [ts_a.key, alpha.key, beta.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    ts_a.key: [1 * S, 2 * S, 3 * S],
                    alpha.key: a1,
                    beta.key: b1,
                }
            )

        # Domain 1 on index B: 2 samples (different count than A)
        g1 = [5.0, 15.0]
        with client.open_writer(
            1 * S, [ts_b.key, gamma.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    ts_b.key: [1 * S, 2 * S],
                    gamma.key: g1,
                }
            )

        # Verify reads after domain 1
        res_a = client.read(sy.TimeRange.MAX, all_a_calc_keys)
        verify_read(res_a, expect_a(a1, b1), "phase1-domain1-A")

        res_b = client.read(sy.TimeRange.MAX, all_b_calc_keys)
        verify_read(res_b, expect_b(g1), "phase1-domain1-B")

        # Domain 2 on index A: 4 samples (different batch size)
        a2, b2 = [100.0, 5.0, 25.0, 40.0], [3.0, 5.0, 2.0, 3.0]
        with client.open_writer(
            10 * S, [ts_a.key, alpha.key, beta.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    ts_a.key: [10 * S, 11 * S, 12 * S, 13 * S],
                    alpha.key: a2,
                    beta.key: b2,
                }
            )

        # Domain 2 on index B: 5 samples (more than A domain 2)
        g2 = [1.0, 100.0, 50.0, 8.0, 200.0]
        with client.open_writer(
            10 * S, [ts_b.key, gamma.key], enable_auto_commit=True
        ) as w:
            w.write(
                {
                    ts_b.key: [10 * S, 11 * S, 12 * S, 13 * S, 14 * S],
                    gamma.key: g2,
                }
            )

        # Full read - should see both domains concatenated
        all_a = a1 + a2
        all_b = b1 + b2
        all_g = g1 + g2
        res_a = client.read(sy.TimeRange.MAX, all_a_calc_keys)
        verify_read(res_a, expect_a(all_a, all_b), "phase1-all-A")
        res_b = client.read(sy.TimeRange.MAX, all_b_calc_keys)
        verify_read(res_b, expect_b(all_g), "phase1-all-B")

        # ── PHASE 2: Streaming with writer lifecycle chaos ────────────────
        # Open streamer on ALL calc channels (both indexes)
        with client.open_streamer(all_a_calc_keys) as streamer_a:
            # Write domain 3 on index A with a fresh writer
            a3, b3 = [8.0, 15.0], [4.0, 1.0]
            with client.open_writer(
                20 * S,
                [ts_a.key, alpha.key, beta.key],
                enable_auto_commit=True,
            ) as w:
                w.write(
                    {
                        ts_a.key: [20 * S, 21 * S],
                        alpha.key: a3,
                        beta.key: b3,
                    }
                )

            frame = streamer_a.read(timeout=5)
            verify_stream_frame(frame, expect_a(a3, b3), "phase2-domain3")

            # Close writer, open a NEW writer, write again
            # This tests that the streamer survives writer lifecycle changes
            a4, b4 = [200.0, 7.0, 33.0], [2.0, 8.0, 4.0]
            with client.open_writer(
                30 * S,
                [ts_a.key, alpha.key, beta.key],
                enable_auto_commit=True,
            ) as w:
                w.write(
                    {
                        ts_a.key: [30 * S, 31 * S, 32 * S],
                        alpha.key: a4,
                        beta.key: b4,
                    }
                )

            frame = streamer_a.read(timeout=5)
            verify_stream_frame(frame, expect_a(a4, b4), "phase2-domain4")

            # Rapid-fire: open writer, write 1 sample, close, repeat 3 times
            rapid_a = []
            rapid_b = []
            for i, (av, bv) in enumerate([(42.0, 6.0), (99.0, 1.0), (3.0, 9.0)]):
                t = (40 + i * 10) * S
                with client.open_writer(
                    t,
                    [ts_a.key, alpha.key, beta.key],
                    enable_auto_commit=True,
                ) as w:
                    w.write(
                        {
                            ts_a.key: [t],
                            alpha.key: [av],
                            beta.key: [bv],
                        }
                    )
                rapid_a.append(av)
                rapid_b.append(bv)

                frame = streamer_a.read(timeout=5)
                verify_stream_frame(frame, expect_a([av], [bv]), f"phase2-rapid-{i}")

        # ── PHASE 3: Close streamer, open new one, keep writing ───────────
        # This tests that a fresh streamer picks up new data correctly
        with client.open_streamer(all_a_calc_keys) as streamer_a2:
            a5, b5 = [11.0, 22.0], [3.0, 6.0]
            with client.open_writer(
                70 * S,
                [ts_a.key, alpha.key, beta.key],
                enable_auto_commit=True,
            ) as w:
                # Write in two separate batches within the same writer
                w.write(
                    {
                        ts_a.key: [70 * S],
                        alpha.key: [a5[0]],
                        beta.key: [b5[0]],
                    }
                )
                frame = streamer_a2.read(timeout=5)
                verify_stream_frame(frame, expect_a([a5[0]], [b5[0]]), "phase3-batch1")

                w.write(
                    {
                        ts_a.key: [71 * S],
                        alpha.key: [a5[1]],
                        beta.key: [b5[1]],
                    }
                )
                frame = streamer_a2.read(timeout=5)
                verify_stream_frame(frame, expect_a([a5[1]], [b5[1]]), "phase3-batch2")

        # ── PHASE 4: Concurrent streamers on different indexes ────────────
        with client.open_streamer(all_b_calc_keys) as streamer_b:
            with client.open_streamer(all_a_calc_keys) as streamer_a3:
                # Write to both indexes simultaneously
                a6, b6 = [55.0], [5.0]
                g3 = [77.0]
                with client.open_writer(
                    80 * S,
                    [ts_a.key, alpha.key, beta.key],
                    enable_auto_commit=True,
                ) as wa:
                    wa.write(
                        {
                            ts_a.key: [80 * S],
                            alpha.key: a6,
                            beta.key: b6,
                        }
                    )

                with client.open_writer(
                    80 * S,
                    [ts_b.key, gamma.key],
                    enable_auto_commit=True,
                ) as wb:
                    wb.write(
                        {
                            ts_b.key: [80 * S],
                            gamma.key: g3,
                        }
                    )

                frame_a = streamer_a3.read(timeout=5)
                verify_stream_frame(frame_a, expect_a(a6, b6), "phase4-A")

                frame_b = streamer_b.read(timeout=5)
                verify_stream_frame(frame_b, expect_b(g3), "phase4-B")

        # ── PHASE 5: Full read of everything, verify total counts ─────────
        # Collect all alpha/beta values written across all domains
        final_a = a1 + a2 + a3 + a4 + rapid_a + a5 + a6
        final_b = b1 + b2 + b3 + b4 + rapid_b + b5 + b6
        final_g = g1 + g2 + g3

        res_a = client.read(sy.TimeRange.MAX, all_a_calc_keys)
        verify_read(res_a, expect_a(final_a, final_b), "phase5-final-A")
        for key in all_a_calc_keys:
            assert len(res_a[key]) == len(final_a), (
                f"Expected {len(final_a)} samples for calc {key}, "
                f"got {len(res_a[key])}"
            )

        res_b = client.read(sy.TimeRange.MAX, all_b_calc_keys)
        verify_read(res_b, expect_b(final_g), "phase5-final-B")
        for key in all_b_calc_keys:
            assert len(res_b[key]) == len(final_g), (
                f"Expected {len(final_g)} samples for calc {key}, "
                f"got {len(res_b[key])}"
            )



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example demonstrates how to export data from a Synnax cluster to a CSV file. We
will write data to a Synnax cluster, and then read the data back and export it to a CSV
file.
"""

import numpy as np

import synnax as sy

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

# Create an index channel that will be used to store our timestamps.
time_channel = client.channels.create(
    name="export_to_csv_time",
    data_type=sy.DataType.TIMESTAMP,
    is_index=True,
)

# Create a data channel that will be used to store our data.
data_channel = client.channels.create(
    name="export_to_csv_data",
    data_type=sy.DataType.FLOAT32,
    index=time_channel.key,
)


# We'll start our write at the current time. This timestamp should be the same as or
# before the first timestamp we write.
start = sy.TimeStamp.now()

# We'll end our write 1 minute later
end = start + sy.TimeSpan.MINUTE

# We'll write 1000 samples of data to Synnax.
SAMPLE_COUNT = 1000

# Generate linearly spaced timestamps
timestamps = np.linspace(start, end, SAMPLE_COUNT)

# Generate linearly spaced data with some noise
data = np.linspace(0, 100, SAMPLE_COUNT) + np.random.randn(SAMPLE_COUNT) * 5

# Write the data to the Synnax cluster through the channels.
client.write(
    start,
    {
        time_channel: timestamps,
        data_channel: data,
    },
)

# Create a time range for reading data.
time_range = sy.TimeRange(start, end)

# Read the data back from the Synnax cluster.
data = client.read(time_range, [time_channel.name, data_channel.name])

# Convert the data to a pandas DataFrame and export to a CSV file.
df = data.to_df().to_csv("exported_data.csv")



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import serial

import synnax as sy

PORT = "/dev/cu.usbmodem21401"
BAUD_RATE = 9600

ser = serial.Serial(PORT, BAUD_RATE)

client = sy.Synnax(
    host="localhost",
    port=9090,
    username="synnax",
    password="seldon",
)

command_channel = client.channels.create(
    name="arduino_command",
    data_type="uint8",
    virtual=True,
    retrieve_if_name_exists=True,
)

with client.open_streamer(["arduino_command"]) as streamer:
    for frame in streamer:
        command = str(frame["arduino_command"][0])
        print(f"Sending command: {command}")
        ser.write(command.encode("utf-8"))



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.


from synnax.access.role.client import Client
from synnax.access.role.payload import ONTOLOGY_TYPE, Role, ontology_id
from synnax.util.deprecation import deprecated_getattr

_DEPRECATED = {
    "RoleClient": "Client",
}

__getattr__ = deprecated_getattr(__name__, _DEPRECATED, globals())

__all__ = ["Role", "Client", "ONTOLOGY_TYPE", "ontology_id"]



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import types
import warnings

import pytest

from synnax.util.deprecation import deprecated_getattr


def _make_module(
    deprecated: dict,
    **globals_entries: object,
) -> types.ModuleType:
    """Create a fake module with deprecated_getattr configured."""
    mod = types.ModuleType("test_module")
    mod.__dict__.update(globals_entries)
    mod.__dict__["__getattr__"] = deprecated_getattr(
        "test_module", deprecated, mod.__dict__
    )
    return mod


@pytest.mark.deprecation
class TestDeprecatedGetattr:
    def test_emits_deprecation_warning(self):
        """Should emit a DeprecationWarning when accessing a deprecated name."""
        mod = _make_module({"OldName": "NewName"}, NewName="value")
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", DeprecationWarning)
            result = mod.OldName
        assert result == "value"
        assert len(w) == 1
        assert issubclass(w[0].category, DeprecationWarning)
        assert "OldName is deprecated, use NewName instead" in str(w[0].message)

    def test_returns_correct_value(self):
        """Should return the same object as the new name."""
        sentinel = object()
        mod = _make_module({"Old": "New"}, New=sentinel)
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always", DeprecationWarning)
            assert mod.Old is sentinel

    def test_caches_after_first_access(self):
        """Should not call __getattr__ on subsequent accesses."""
        mod = _make_module({"Old": "New"}, New="value")
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always", DeprecationWarning)
            _ = mod.Old
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", DeprecationWarning)
            _ = mod.Old
        assert len(w) == 0

    def test_raises_attribute_error_for_unknown(self):
        """Should raise AttributeError for names that are not deprecated."""
        mod = _make_module({}, NewName="value")
        with pytest.raises(AttributeError, match="test_module"):
            _ = mod.NonExistent

    def test_tuple_form_custom_display_name(self):
        """Should use the display name from a tuple entry in the warning."""
        mod = _make_module(
            {"OldName": ("package.module.NewName", "_internal")},
            _internal="value",
        )
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", DeprecationWarning)
            result = mod.OldName
        assert result == "value"
        assert len(w) == 1
        assert "use package.module.NewName instead" in str(w[0].message)

    def test_tuple_form_caches(self):
        """Should cache after first access with tuple form."""
        mod = _make_module(
            {"OldName": ("pkg.New", "_internal")},
            _internal="value",
        )
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always", DeprecationWarning)
            _ = mod.OldName
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", DeprecationWarning)
            _ = mod.OldName
        assert len(w) == 0

    def test_multiple_deprecated_names(self):
        """Should handle multiple deprecated names independently."""
        mod = _make_module(
            {"OldA": "NewA", "OldB": "NewB"},
            NewA="a",
            NewB="b",
        )
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", DeprecationWarning)
            assert mod.OldA == "a"
            assert mod.OldB == "b"
        assert len(w) == 2
        assert "OldA" in str(w[0].message)
        assert "OldB" in str(w[1].message)

    def test_non_deprecated_access_unaffected(self):
        """Should not interfere with normal attribute access."""
        mod = _make_module({"Old": "New"}, New="value", Other="other")
        assert mod.Other == "other"
        assert mod.New == "value"



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from synnax.framer.client import Client
from synnax.framer.deleter import Deleter
from synnax.framer.frame import CrudeFrame, Frame
from synnax.framer.iterator import AUTO_SPAN, Iterator
from synnax.framer.streamer import AsyncStreamer, Streamer
from synnax.framer.writer import Writer, WriterMode

__all__ = [
    "Client",
    "Deleter",
    "Frame",
    "CrudeFrame",
    "AUTO_SPAN",
    "Iterator",
    "AsyncStreamer",
    "Streamer",
    "Writer",
    "WriterMode",
]



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from pathlib import Path

from .protocol import Matcher


def new_extension_matcher(extensions: list[str]) -> type[Matcher]:
    """Return a matcher that matches file extensions.

    :returns: a matcher that matches file extensions.
    """
    _e = extensions

    class ExtensionMatcher:
        _extensions = _e

        @classmethod
        def match(cls, path: Path) -> bool:
            return path.suffix[1:] in cls.extensions()

        @classmethod
        def extensions(cls) -> list[str]:
            return cls._extensions

    return ExtensionMatcher



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import time

import numpy as np
import pytest

import synnax as sy


@pytest.mark.timing
class TestTiming:
    def test_sleep(self):
        """Test that the sleep function is consistently better that time.sleep in terms
        of precision.

        Execute 50 different timing tests at random intervals between 100 microseconds
        and 5 milliseconds with both time.sleep and sy.sleep.
        """
        accumulated_precise = list()
        accumulated_standard = list()
        for _ in range(50):
            duration = sy.TimeSpan.MICROSECOND * float(np.random.uniform(100, 5_000))
            start = time.perf_counter_ns()
            sy.sleep(duration.seconds, precise=False)
            time_elapsed = time.perf_counter_ns() - start
            start = time.perf_counter_ns()
            sy.sleep(duration, precise=True)
            sy_elapsed = time.perf_counter_ns() - start
            standard_delta = abs(time_elapsed - duration.nanoseconds)
            precise_delta = abs(sy_elapsed - duration.nanoseconds)
            accumulated_precise.append(precise_delta)
            accumulated_standard.append(standard_delta)

        assert sum(accumulated_precise) < sum(accumulated_standard)

    def test_sleep_rate(self):
        """Should sleep correctly based on a rate argument"""
        t = sy.Timer()
        sy.sleep(100 * sy.Rate.HZ, precise=True)
        print((100 * sy.Rate.HZ).period)
        assert t.elapsed() < sy.TimeSpan.MILLISECOND * 11
        assert t.elapsed() > sy.TimeSpan.MILLISECOND * 9

    def test_loop(self):
        """Test that the loop holds timing consistent even when operations in the loop
        take a long time.
        """
        loop = sy.Loop(sy.TimeSpan.MILLISECOND * 10, precise=True)
        i = 0
        start = time.perf_counter_ns()
        with loop:
            for _ in loop:
                i += 1
                if i == 10:
                    break
                sy.sleep(sy.TimeSpan.MILLISECOND * 5, precise=True)
        end = time.perf_counter_ns()
        assert sy.TimeSpan(end - start) < sy.TimeSpan.MILLISECOND * 110



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import overload

from freighter import Empty, UnaryClient, send_required
from pydantic import BaseModel, Field

from synnax.ontology.payload import ID, CrudeID, Resource
from synnax.util.normalize import normalize


class RetrieveReq(BaseModel):
    ids: list[ID]
    children: bool = False
    parents: bool = False
    include_schema: bool = False
    exclude_field_data: bool = False
    term: str | None = None
    limit: int | None = None
    offset: int | None = None
    types: list[str] | None = None


class AddChildrenReq(BaseModel):
    id: ID
    children: list[ID]


class RemoveChildrenReq(BaseModel):
    id: ID
    children: list[ID]


class MoveChildrenReq(BaseModel):
    from_: ID = Field(alias="from")
    to: ID
    children: list[ID]


class RetrieveRes(BaseModel):
    resources: list[Resource]


class Client:
    _client: UnaryClient

    def __init__(self, client: UnaryClient) -> None:
        self._client = client

    @overload
    def retrieve(
        self,
        id: CrudeID,
        *,
        children: bool = False,
        parents: bool = False,
        include_schema: bool = False,
        exclude_field_data: bool = False,
    ) -> Resource: ...

    @overload
    def retrieve(
        self,
        id: list[CrudeID],
        *,
        children: bool = False,
        parents: bool = False,
        include_schema: bool = False,
        exclude_field_data: bool = False,
    ) -> list[Resource]: ...

    def retrieve(
        self,
        id: CrudeID | list[CrudeID] | None = None,
        *,
        children: bool = False,
        parents: bool = False,
        include_schema: bool = False,
        exclude_field_data: bool = False,
    ) -> Resource | list[Resource]:
        is_single = False
        if not isinstance(id, list):
            if id is None:
                id = []
            else:
                id = [id]
                is_single = True
        req = RetrieveReq(
            ids=[ID(i) for i in id],
            children=children,
            parents=parents,
            include_schema=include_schema,
            exclude_field_data=exclude_field_data,
        )
        resources = self.__exec_retrieve(req)
        if is_single:
            return resources[0]
        return resources

    def retrieve_children(
        self,
        id: CrudeID | list[CrudeID],
    ) -> list[Resource]:
        normalized: list[CrudeID] = normalize(id)
        return self.__exec_retrieve(
            RetrieveReq(ids=[ID(i) for i in normalized], children=True)
        )

    def __exec_retrieve(self, req: RetrieveReq) -> list[Resource]:
        return send_required(
            self._client, "/ontology/retrieve", req, RetrieveRes
        ).resources

    def retrieve_parents(
        self,
        id: CrudeID | list[CrudeID],
    ) -> list[Resource]:
        normalized: list[CrudeID] = normalize(id)
        return self.__exec_retrieve(
            RetrieveReq(ids=[ID(i) for i in normalized], parents=True)
        )

    def move_children(self, from_: CrudeID, to: CrudeID, *children: CrudeID) -> None:

        send_required(
            self._client,
            "/ontology/move-children",
            MoveChildrenReq.model_validate(
                {"from": ID(from_), "to": ID(to), "children": [ID(i) for i in children]}
            ),
            Empty,
        )

    def remove_children(self, id: CrudeID, *children: CrudeID) -> None:
        send_required(
            self._client,
            "/ontology/remove-children",
            RemoveChildrenReq(id=ID(id), children=[ID(i) for i in children]),
            Empty,
        )

    def add_children(self, id: CrudeID, *children: CrudeID) -> None:
        send_required(
            self._client,
            "/ontology/add-children",
            AddChildrenReq(id=ID(id), children=[ID(i) for i in children]),
            Empty,
        )



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example demonstrates task management features in Synnax.

The principles shown here apply to ALL hardware task types (NI, LabJack, OPC UA, Modbus, etc.),
but this example uses Modbus for demonstration purposes.

Key features demonstrated:
- Creating tasks
- Retrieving tasks by name
- Listing all tasks on a rack
- Copying tasks
- Modifying task configurations
- Starting/stopping tasks

Before running this example:
1. Start the Modbus simulator server:
   uv run python driver/modbus/dev/server.py

2. Connect the Modbus device in Synnax Console:
   - Host: localhost (127.0.0.1)
   - Port: 5020
   - Name: "Modbus Server"

Note: The task management API is identical for all hardware types - only the task
configuration specifics differ between hardware integrations.
"""

import synnax as sy
from synnax import modbus

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

# Retrieve the Modbus device from Synnax
# Update this with the name you gave the device in the Synnax Console
dev = client.devices.retrieve(name="Modbus Server")

print("=" * 70)
print("Modbus Task Management Example")
print()

# ============================================================================
# Step 1: Create channels and an initial Modbus read task
# ============================================================================
print("Step 1: Creating initial Modbus read task")
print("-" * 70)

# Create an index channel for timestamps
modbus_time = client.channels.create(
    name="modbus_mgmt_time",
    is_index=True,
    data_type=sy.DataType.TIMESTAMP,
    retrieve_if_name_exists=True,
)

# Create data channels for two input registers
input_reg_0 = client.channels.create(
    name="modbus_mgmt_reg_0",
    index=modbus_time.key,
    data_type=sy.DataType.UINT8,
    retrieve_if_name_exists=True,
)

input_reg_1 = client.channels.create(
    name="modbus_mgmt_reg_1",
    index=modbus_time.key,
    data_type=sy.DataType.UINT8,
    retrieve_if_name_exists=True,
)

# Create the initial Modbus read task
original_task = sy.modbus.ReadTask(
    name="Modbus Read Task",
    device=dev.key,
    sample_rate=sy.Rate.HZ * 10,
    stream_rate=sy.Rate.HZ * 10,
    data_saving=True,
    channels=[
        sy.modbus.InputRegisterChan(
            channel=input_reg_0.key, address=0, data_type="uint8"
        ),
        sy.modbus.InputRegisterChan(
            channel=input_reg_1.key, address=1, data_type="uint8"
        ),
    ],
)

# Configure the task with Synnax
client.tasks.configure(original_task)
print(f"✓ Created task: '{original_task.name}'")
print(f"  Task key: {original_task.key}")
print(f"  Sample rate: 10 Hz")
print(f"  Stream rate: 10 Hz")
print(f"  Channels: 2 input registers")
print()

# ============================================================================
# Step 2: Copy the task and modify its configuration
# ============================================================================
print("\nStep 2: Copying task and modifying configuration")
print("-" * 70)

# Retrieve the original task by name
retrieved_task = client.tasks.retrieve(name="Modbus Read Task")
print(f"\nRetrieved task by name: '{retrieved_task.name}' (key: {retrieved_task.key})")

# Copy the original task with a new name
# This creates a new, independent task with the same configuration
copied_task_raw = client.tasks.copy(
    key=retrieved_task.key,
    name="Modbus Read Task Copy",
)

# Convert the raw task to a ReadTask object so we can modify the configuration
copied_task = sy.modbus.ReadTask(internal=copied_task_raw)

# Modify the stream rate and enable auto-start
# Convert Rate to int to avoid Pydantic serialization warnings
copied_task.config.stream_rate = int(sy.Rate.HZ * 5)
copied_task.config.auto_start = True

# Reconfigure the task with the new settings
# This will also start the task automatically since auto_start is True
client.tasks.configure(copied_task)

print(f"\n✓ Copied and modified task: '{copied_task.name}'")
print(f"  Original task:")
print(f"    Key:          {original_task.key}")
print(f"    Stream rate:  10 Hz")
print(f"    Auto-start:   Disabled")
print(f"  Copied task:")
print(f"    Key:          {copied_task.key}")
print(f"    Stream rate:  5 Hz (modified)")
print(f"    Auto-start:   Enabled (task is now running)")

# Keep the copied task running for now - we'll verify it below
print()

# ============================================================================
# Step 3: List all tasks to see the original and copy
# ============================================================================
print("\nStep 3: Listing all tasks on the rack")
print("-" * 70)

# List all tasks again to see the original and copy
all_tasks_updated = client.tasks.list()
print(f"✓ Found {len(all_tasks_updated)} task(s) on the rack:\n")

# Group tasks by type for better visualization
modbus_tasks = [t for t in all_tasks_updated if t.type == "modbus_read"]
other_tasks = [t for t in all_tasks_updated if t.type != "modbus_read"]

if modbus_tasks:
    print("  Modbus Read Tasks:")
    for task in modbus_tasks:
        print(f"    • {task.name}")
        print(f"      Key: {task.key}")

if other_tasks:
    print("\n  Other Tasks:")
    for task in other_tasks:
        print(f"    • {task.name}")
        print(f"      Key: {task.key}, Type: {task.type}")

print()

# Verify the copied task is running by reading some data
print("  Verifying copied task is running (5 second test)...")
with client.open_streamer([input_reg_0.key, input_reg_1.key]) as streamer:
    start_time = sy.TimeStamp.now()
    sample_count = 0
    while sy.TimeStamp.now().span(start_time).seconds < 5:
        frame = streamer.read()
        if frame and input_reg_0.key in frame:
            sample_count += len(frame[input_reg_0.key])

if sample_count > 0:
    print(f"  ✓ Copied task is running (received {sample_count} samples)")
else:
    print(f"  ✗ No samples received from copied task")

# Stop the copied task to clean up
print("  Stopping copied task...")
copied_task.stop()
print(f"  ✓ Copied task stopped")

# ============================================================================
# Step 4: Delete the copied task
# ============================================================================
print("\nStep 4: Deleting the copied task")
print("-" * 70)

# Delete the copied task by key
client.tasks.delete(copied_task.key)
print(f"✓ Deleted copied task (key: {copied_task.key})")

# Verify deletion by listing tasks again
remaining_tasks = client.tasks.list()
remaining_modbus = [t for t in remaining_tasks if t.type == "modbus_read"]
print(f"✓ Remaining Modbus tasks: {len(remaining_modbus)}")
for task in remaining_modbus:
    print(f"  • {task.name} (key: {task.key})")

print("\n✓ Example completed successfully!")
print("=" * 70)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example demonstrates how to configure and use a write task to send commands
to an OPC UA server.

Before running this example:
1. Start the OPC UA test server:
   uv run python -m examples.opcua.server

2. Connect the OPC UA server device in Synnax:
   uv run python examples/opcua/connect_server.py

3. The server will print the node IDs for the command variables on startup.
   Update the node_id values below with the actual IDs from your server output.

Use --tls to target the TLS-encrypted server instead.
"""

import argparse
import time

import synnax as sy

parser = argparse.ArgumentParser(description="Write commands to OPC UA server")
parser.add_argument(
    "--tls", action="store_true", help="Target the TLS-encrypted server"
)
parser.add_argument(
    "--tls-auth",
    action="store_true",
    help="Target the TLS-encrypted server with username/password (port 4843)",
)
args = parser.parse_args()

if args.tls_auth:
    DEVICE_NAME = "OPC UA TLS Auth Server"
elif args.tls:
    DEVICE_NAME = "OPC UA TLS Server"
else:
    DEVICE_NAME = "OPC UA Server"

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

# Retrieve the OPC UA server from Synnax
dev = client.devices.retrieve(name=DEVICE_NAME)

# Create an index channel for the command channels
opcua_cmd_time = client.channels.create(
    name="opcua_cmd_time",
    is_index=True,
    data_type=sy.DataType.TIMESTAMP,
    retrieve_if_name_exists=True,
)

# Create command channels that will be used to send control values to the OPC UA nodes.
# These are regular Synnax channels - the write task will listen to them and forward
# the values to the OPC UA server.
cmd_channel_0 = client.channels.create(
    name="opcua_cmd_0",
    index=opcua_cmd_time.key,
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
)
cmd_channel_1 = client.channels.create(
    name="opcua_cmd_1",
    index=opcua_cmd_time.key,
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
)
cmd_channel_2 = client.channels.create(
    name="opcua_cmd_2",
    index=opcua_cmd_time.key,
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
)

# Create and configure the OPC UA write task
# NOTE: Update these node IDs with the actual values from your server output!
# The server prints them on startup like: "command_0: ns=2;i=42"
tsk = sy.opcua.WriteTask(
    name="OPC UA Write Task Example",
    device=dev.key,
    auto_start=False,  # We'll start it manually for this example
    channels=[
        # Map Synnax command channels to OPC UA node IDs
        # Replace these node_id values with the actual IDs from your server
        sy.opcua.WriteChannel(cmd_channel=cmd_channel_0.key, node_id="NS=2;I=18"),
        sy.opcua.WriteChannel(cmd_channel=cmd_channel_1.key, node_id="NS=2;I=19"),
        sy.opcua.WriteChannel(cmd_channel=cmd_channel_2.key, node_id="NS=2;I=20"),
    ],
)

# Configure the task with Synnax
client.tasks.configure(tsk)

print("=" * 70)
print("Starting OPC UA Write Task")
print("=" * 70)
print("Sending commands to the OPC UA server...")
print("Writing 10 commands at 1 Hz (10 seconds)")

print(f"{'Cycle':<8} {'command_0':>12} {'command_1':>12} {'command_2':>12}")
print("-" * 70)

# Start the write task
with tsk.run():
    # Open a writer to send commands to the OPC UA server
    # IMPORTANT: Must include the index channel (opcua_cmd_time) along with data channels
    with client.open_writer(
        start=sy.TimeStamp.now(),
        channels=[
            opcua_cmd_time.key,
            cmd_channel_0.key,
            cmd_channel_1.key,
            cmd_channel_2.key,
        ],
        enable_auto_commit=True,
    ) as writer:
        # Write some test values to the OPC UA nodes
        for i in range(10):
            val0 = float(i)
            val1 = float(i * 2)
            val2 = float(i * 3)

            print(f"{i+1:<8} {val0:>12.1f} {val1:>12.1f} {val2:>12.1f}")

            # Write timestamp to index channel along with command values
            writer.write(
                {
                    opcua_cmd_time.key: sy.TimeStamp.now(),
                    cmd_channel_0.key: val0,
                    cmd_channel_1.key: val1,
                    cmd_channel_2.key: val2,
                }
            )
            writer.commit()
            sy.sleep(1)

print("-" * 70)
print("✓ Write task completed successfully!")
print("Check your server terminal to verify the commands were received:")
print("You should see timestamps and values for each command variable")



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example demonstrates how to read boolean data from the OPC UA test server.

Before running this example:
1. Start the test server:
   uv run python -m examples.opcua.server

2. Connect the OPC UA server device in Synnax:
   uv run python examples/opcua/connect_server.py

3. The server creates boolean variables (my_bool_0, my_bool_1, etc.) that continuously
   update with sequential square wave patterns. This example reads those values.

Use --tls to target the TLS-encrypted server instead.
"""

import argparse

import synnax as sy

parser = argparse.ArgumentParser(description="Read boolean data from OPC UA server")
parser.add_argument(
    "--tls", action="store_true", help="Target the TLS-encrypted server"
)
parser.add_argument(
    "--tls-auth",
    action="store_true",
    help="Target the TLS-encrypted server with username/password (port 4843)",
)
args = parser.parse_args()

if args.tls_auth:
    DEVICE_NAME = "OPC UA TLS Auth Server"
elif args.tls:
    DEVICE_NAME = "OPC UA TLS Server"
else:
    DEVICE_NAME = "OPC UA Server"

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

# Retrieve the OPC UA server from Synnax
dev = client.devices.retrieve(name=DEVICE_NAME)

# Create an index channel that will be used to store the timestamps for the data.
opcua_bool_time = client.channels.create(
    name="opcua_bool_time",
    is_index=True,
    data_type=sy.DataType.TIMESTAMP,
    retrieve_if_name_exists=True,
)

# Create Synnax channels to store the my_bool_0 and my_bool_1 node data from the server.
# These booleans are continuously updated with sequential square wave patterns by the OPC UA server.
my_bool_0 = client.channels.create(
    name="my_bool_0",
    index=opcua_bool_time.key,
    data_type=sy.DataType.UINT8,  # Boolean data is stored as UINT8 (0 or 1)
    retrieve_if_name_exists=True,
)
my_bool_1 = client.channels.create(
    name="my_bool_1",
    index=opcua_bool_time.key,
    data_type=sy.DataType.UINT8,  # Boolean data is stored as UINT8 (0 or 1)
    retrieve_if_name_exists=True,
)

# Create the OPC UA Read Task
# Using node IDs to reference the OPC UA boolean variables.
# The server creates booleans with sequential square wave patterns offset by 0.2 seconds.
tsk = sy.opcua.ReadTask(
    name="OPC UA Py - Read Task (Boolean)",
    device=dev.key,
    sample_rate=sy.Rate.HZ * 10,  # Sample at 10 Hz
    stream_rate=sy.Rate.HZ * 10,  # Stream at 10 Hz
    data_saving=True,
    channels=[
        # Bind the Synnax channels to the OPC UA node IDs
        # These IDs correspond to my_bool_0 and my_bool_1 in the OPC UA server
        sy.opcua.ReadChannel(
            channel=my_bool_0.key, node_id="NS=2;I=13", data_type="bool"  # my_bool_0
        ),
        sy.opcua.ReadChannel(
            channel=my_bool_1.key, node_id="NS=2;I=14", data_type="bool"  # my_bool_1
        ),
    ],
)

# Configure the task with Synnax
client.tasks.configure(tsk)

print("=" * 70)
print("Starting OPC UA Boolean Read Task")
print("=" * 70)
print("Reading square wave boolean data from the OPC UA server...")
print("Running continuously - Press Ctrl+C to stop\n")

print(f"{'Sample':<8} {'Timestamp':<12} {'my_bool_0':>12} {'my_bool_1':>12}")
print("-" * 70)

# Start the task and read data continuously
try:
    # Hide cursor for clean output
    print("\033[?25l", end="", flush=True)

    with tsk.run():
        with client.open_streamer(["my_bool_0", "my_bool_1"]) as streamer:
            sample_count = 0
            start_time = sy.TimeStamp.now()

            while True:
                frame = streamer.read()
                if frame:
                    # Print the latest values from each channel
                    if "my_bool_0" in frame and len(frame["my_bool_0"]) > 0:
                        val0 = frame["my_bool_0"][-1]
                        val1 = frame["my_bool_1"][-1]

                        elapsed = sy.TimeStamp.now().span(start_time).seconds

                        sample_count += 1
                        # Convert uint8 to True/False for display
                        bool0 = "True " if val0 else "False"
                        bool1 = "True " if val1 else "False"
                        print(
                            f"{sample_count:<8} {elapsed:<12.1f} {bool0:>12} {bool1:>12}",
                            end="\r",
                            flush=True,
                        )

# Output summary
except KeyboardInterrupt:
    print("\n" + "-" * 70)
    print("✓ Boolean read task stopped by user")
    print(f"\nCollected {sample_count} samples")
    print("The values are sequential square waves:")
    print("- my_bool_0: toggles every second starting at t=0")
    print("- my_bool_1: toggles every second starting at t=0.2")
    print("=" * 70)
finally:
    # Ensure cursor is always shown even if something goes wrong
    print("\033[?25h", end="", flush=True)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.


from freighter import URL, UnaryClient
from freighter import send_required as fsend_required
from freighter.exceptions import Unreachable
from freighter.transport import RQ, RS


def send_required(client: UnaryClient, target: str, req: RQ, res_t: type[RS]) -> RS:
    """Sends a request to a target and returns the response. Raises an exception if the
    request returns an error.
    """
    try:
        return fsend_required(client, target, req, res_t)
    except Unreachable as exc:
        url = URL.parse(exc.target)
        raise Unreachable(
            target=url.stringify(),
            message=f"Failed to reach cluster at {url.host}:{url.port}",
        ) from exc



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import overload
from uuid import UUID

from alamos import NOOP, Instrumentation
from freighter import Empty, UnaryClient, send_required
from pydantic import BaseModel

from synnax.access.policy.payload import Policy
from synnax.ontology.payload import ID
from synnax.util.normalize import normalize


class _CreateRequest(BaseModel):
    policies: list[Policy]


_CreateResponse = _CreateRequest


class _RetrieveRequest(BaseModel):
    keys: list[UUID] | None
    subjects: list[ID] | None
    internal: bool | None = None


class _RetrieveResponse(BaseModel):
    policies: list[Policy] | None


class _DeleteRequest(BaseModel):
    keys: list[UUID]


ONTOLOGY_TYPE = ID(type="policy")


class Client:
    _client: UnaryClient
    instrumentation: Instrumentation

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ):
        self._client = client
        self.instrumentation = instrumentation

    @overload
    def create(
        self,
        policies: Policy,
    ) -> Policy: ...

    @overload
    def create(
        self,
        policies: list[Policy],
    ) -> list[Policy]: ...

    def create(
        self,
        policies: Policy | list[Policy],
    ) -> Policy | list[Policy]:
        is_single = not isinstance(policies, list)
        req = _CreateRequest(policies=normalize(policies))
        res = send_required(self._client, "/access/policy/create", req, _CreateResponse)
        return res.policies[0] if is_single else res.policies

    def retrieve(
        self,
        keys: list[UUID] | None = None,
        subjects: list[ID] | None = None,
        internal: bool | None = None,
    ) -> list[Policy]:
        res = send_required(
            self._client,
            "/access/policy/retrieve",
            _RetrieveRequest(keys=keys, subjects=subjects, internal=internal),
            _RetrieveResponse,
        )
        return [] if res.policies is None else res.policies

    def delete(self, keys: UUID | list[UUID]) -> None:
        req = _DeleteRequest(keys=normalize(keys))
        send_required(self._client, "/access/policy/delete", req, Empty)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import Any, Generic, Literal, TypeVar
from uuid import uuid4

from pydantic import BaseModel, Field

from synnax.ontology import ID
from synnax.telem import TimeStamp

VARIANT_SUCCESS = "success"
VARIANT_INFO = "info"
VARIANT_WARNING = "warning"
VARIANT_ERROR = "error"
VARIANT_DISABLED = "disabled"
VARIANT_LOADING = "loading"

Variant = Literal[
    "success",
    "info",
    "warning",
    "error",
    "disabled",
    "loading",
]
"""Represents the variant of a status message."""

D = TypeVar("D", bound=BaseModel | None)

ONTOLOGY_TYPE = ID(type="status")


def ontology_id(key: str) -> ID:
    """Create an ontology ID for a status.

    Args:
        key: The status key.

    Returns:
        An ontology ID dictionary with type "status" and the given key.
    """
    return ID(type=ONTOLOGY_TYPE.type, key=key)


class Status(BaseModel, Generic[D]):
    """A standardized payload used across Synnax."""

    key: str = Field(default_factory=lambda: str(uuid4()))
    """A unique key for the status."""
    name: str = ""
    """A human-readable name for the status."""
    variant: Variant
    """The variant of the status."""
    message: str
    """The message of the status."""
    description: str = ""
    """The description of the status."""
    time: TimeStamp = Field(default_factory=TimeStamp.now)
    """The time the status was created."""
    labels: list[Any] | None = None
    """Optional labels attached to the status (only present in responses)."""
    details: D | None = None
    """The details are customizable details for component specific statuses."""

    @property
    def ontology_id(self) -> ID:
        """Get the ontology ID for the status.

        Returns:
            An ontology ID dictionary with type "status" and the status key.
        """
        return ontology_id(self.key)


from synnax.util.deprecation import deprecated_getattr

_DEPRECATED = {
    "SUCCESS_VARIANT": "VARIANT_SUCCESS",
    "INFO_VARIANT": "VARIANT_INFO",
    "WARNING_VARIANT": "VARIANT_WARNING",
    "ERROR_VARIANT": "VARIANT_ERROR",
    "DISABLED_VARIANT": "VARIANT_DISABLED",
    "LOADING_VARIANT": "VARIANT_LOADING",
}

__getattr__ = deprecated_getattr(__name__, _DEPRECATED, globals())



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from alamos import NOOP, Instrumentation, trace
from freighter import Empty, UnaryClient, send_required
from pydantic import BaseModel

from synnax.ontology.payload import ID
from synnax.ranger.payload import Key, Payload


class _CreateRequest(BaseModel):
    parent: ID | None = None
    ranges: list[Payload] = []


_CreateResponse = _CreateRequest


class _DeleteRequest(BaseModel):
    keys: list[Key]


class Writer:
    _client: UnaryClient
    instrumentation: Instrumentation

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self._client = client
        self.instrumentation = instrumentation

    @trace("debug", "range.create")
    def create(
        self, ranges: list[Payload], *, parent: ID | None = None
    ) -> list[Payload]:
        req = _CreateRequest(ranges=ranges, parent=parent)
        return send_required(self._client, "/range/create", req, _CreateResponse).ranges

    @trace("debug", "range.delete")
    def delete(self, keys: list[Key]) -> None:
        req = _DeleteRequest(keys=keys)
        send_required(self._client, "/range/delete", req, Empty)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.


from typing import Any

from pydantic import BaseModel

from synnax import ontology, status

Key = str

ONTOLOGY_TYPE = ontology.ID(type="device")


def ontology_id(key: str) -> ontology.ID:
    """Returns the ontology ID for the Device entity."""
    return ontology.ID(type=ONTOLOGY_TYPE.type, key=key)


class StatusDetails(BaseModel):
    """Details about the status of a device."""

    rack: int = 0
    """The key of the rack the device is connected to."""
    device: str = ""
    """The key of the device."""


Status = status.Status[StatusDetails]
"""The status of a device."""


class Device(BaseModel):
    key: str = ""
    location: str = ""
    rack: int = 0
    name: str = ""
    make: str = ""
    model: str = ""
    configured: bool = False
    properties: dict[str, Any] = {}
    status: Status | None = None

    @property
    def ontology_id(self) -> ontology.ID:
        """Returns the ontology ID for this Device."""
        return ontology_id(self.key)
